#    Copyright 2022 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
import json
from datetime import datetime, timedelta

import pytest

from si_tests import logger, settings
from si_tests.deployments.deploy_refapp import Refapp
from si_tests.managers.bootstrap_manager import BootstrapManager
from si_tests.managers.load_tools_manager import LoadToolsManager
from si_tests.managers.openstack_client_manager import OpenStackClientManager

LOG = logger.logger


def calculate_loadtools_downtime(data: dict) -> list:
    """
    Calculate potential downtime period

    Args:
        data: LoadToolsManager.get_custom_statistics object

    Returns: result list of downtimes

    """
    result = []
    start = {}
    stop = {}
    for item in data:
        if item['fail_per_sec']:
            if not start:
                start = item
            else:
                stop = item
        else:
            if start and stop:
                time_format = '%H:%M:%S'
                start_data = datetime.strptime(start['time_data'], time_format)
                stop_data = datetime.strptime(stop['time_data'], time_format)
                if stop_data < start_data:
                    # workaround to handle next day because only time is stored
                    stop_data += timedelta(days=1)
                duration = stop_data - start_data

                result.append({
                    "start": start_data.strftime(time_format),
                    "stop": stop_data.strftime(time_format),
                    "duration": duration
                })
            start = {}
            stop = {}
    return result


@pytest.fixture(scope='function')
def loadtest_os_refapp(kaas_manager):
    """
    Deploy RefApp to OpenStack Cluster (if not exist) as a workload

    """
    if settings.ALLOW_WORKLOAD:
        LOG.info("Start RefApp load")
        managed_ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
        cluster = managed_ns.get_cluster(settings.TARGET_CLUSTER)

        child_kubeconfig_name, child_kubeconfig = \
            cluster.get_kubeconfig_from_secret()
        with open(child_kubeconfig_name, 'w') as f:
            f.write(child_kubeconfig)
        cluster_os_manager = OpenStackClientManager(kubeconfig=child_kubeconfig_name)

        refapp = Refapp(cluster_os_manager, kubeconfig=child_kubeconfig_name)
        refapp.deploy()
        LOG.info(f"RefApp URL = {refapp.url}")

        if settings.SEED_STANDALONE_EXTERNAL_IP:
            seed_ip = settings.SEED_STANDALONE_EXTERNAL_IP
        else:
            seed_ip = kaas_manager.si_config.data.get('run_on_remote', {}).get('SEED_STANDALONE_EXTERNAL_IP')
        assert seed_ip, 'Seed node IP address not found'
        LOG.info(f"Seed node IP address {seed_ip}")
        bootstrap_manager = BootstrapManager(seed_ip=seed_ip)
        load_manager = LoadToolsManager(bootstrap_manager=bootstrap_manager, refapp_address=refapp.url)
        load_manager.run_smoke()
        load_manager.run_loadtools_on_seed_node()
        load_manager.start_swarm()

        yield

        try:

            datetime_prefix = datetime.now().strftime("%Y%m%d%H%M%S")
            load_manager.stop_swarm()
            statistics = load_manager.get_custom_statistics()

            json_statistic_filename = f"statistics-{cluster.name}-{datetime_prefix}.json"

            with open(f'{settings.ARTIFACTS_DIR}{json_statistic_filename}', 'w') as fp:
                json.dump(statistics, fp)

            LOG.info(f"Save JSON statistic to artifacts/{json_statistic_filename}")

            LOG.info("Check ReffApp downtime")

            downtime = calculate_loadtools_downtime(statistics)
            cluster.check.check_downtime(downtime=downtime)

            LOG.info("Check RefApp after test")
            load_manager.run_smoke()
        except Exception as e:
            raise e
        finally:
            load_manager.exit()
    else:
        LOG.info("Skip RefApp load")

        yield
