import random
from datetime import datetime, timezone, timedelta

import dateutil.parser
import exec_helpers
import yaml
import time
import json
import kubernetes
import socket
import re
import pytest

from retry import retry
from si_tests.utils import packaging_version as version
from tabulate import tabulate
from urllib3.exceptions import MaxRetryError, ProtocolError
from si_tests import settings
from si_tests import logger
from si_tests.utils import utils, waiters, exceptions, templates, helpers
from kubernetes.client.rest import ApiException
from websocket._exceptions import WebSocketBadStatusException
from OpenSSL.SSL import Connection, Context, SSLv23_METHOD, TLSv1_2_METHOD
from si_tests.fixtures.cluster import collect_cluster_readiness

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from si_tests.managers.kaas_manager import Cluster, Machine

LOG = logger.logger


class ClusterCheckManager(object):
    """Cluster check manager"""

    EXCLUDED_PODS = [
        'elasticsearch-curator',
        'ui-e2e-test',
        # UCP internal pods
        'ucp-nvidia-gpu-feature-discovery',
        # compose and compose-api are deprecated
        'compose',
        'compose-api',
        'tungsten-pytest'
    ]
    EXCLUDED_JOBS = [
        'elasticsearch-curator',
        'mariadb-phy-backup'
    ]
    EXCLUDED_ROLESBINDS = []

    def __init__(self, cluster: "Cluster"):
        self._cluster: "Cluster" = cluster
        self.deploys_generation = {}
        self.deploys_next_generation = {}

    @property
    def cluster(self) -> "Cluster":
        return self._cluster

    @property
    def k8sclient(self):
        return self._cluster.k8sclient

    def get_deploys_generation(self, namespace=None, deploys=None):
        deploys_generation = {}
        if not deploys and not namespace:
            deploys = self.cluster.k8sclient.deployments.list_all()
        if namespace and not deploys:
            deploys = [d for d in self.cluster.k8sclient.deployments.list(
                namespace=namespace)]
        for deploy in deploys:
            deploy_status = deploy.read().to_dict().get('status') or {}
            generation = deploy_status.get('observed_generation') or 0
            deploys_generation[deploy.name] = int(generation)
        return deploys_generation

    def store_deploys_generation(self, namespace=None, deploys=None):
        generations = self.get_deploys_generation(
            namespace=namespace, deploys=deploys)
        self.deploys_generation = generations
        self.deploys_next_generation = {k: v+1 for k, v in generations.items()}

    def wait_deploy_generation_changed(self, namespace=None, deploys=None,
                                       timeout=600, interval=60):
        LOG.info("Waiting for generations are changed")
        for k, v in self.deploys_generation.items():
            LOG.info("{}: {} -> {}".format(
                k, v, self.deploys_next_generation[k]))

        waiters.wait(lambda: self.get_deploys_generation(
            namespace=namespace,
            deploys=deploys) == self.deploys_next_generation,
            timeout=timeout, interval=interval,
            timeout_msg="Some generations are not changed. Current "
                        "generations info: \n{} but should be: \n{}".format(
            yaml.dump(self.get_deploys_generation(namespace=namespace,
                                                  deploys=deploys)),
            yaml.dump(self.deploys_next_generation)))

    def get_nodeworkloadlock_state(self, name):
        workloadlock = self.cluster.get_nodeworkloadlock(name)
        if not workloadlock:
            raise ValueError(f"Node workloadlock with name {name} not "
                             f"found")
        workloadlock_status = workloadlock.data.get('status') or {}
        workloadlock_state = workloadlock_status.get('state')
        return workloadlock_state

    def wait_nodeworkloadlock_state(self, name, state='active',
                                    timeout=2400, interval=60):

        def status_message():
            info = dict()
            info['expected_state'] = state
            workloadlock = self.cluster.get_nodeworkloadlock(name)
            workloadlock_status = workloadlock.data.get('status') or {}
            workloadlock_state = workloadlock_status.get('state')
            info['current_state'] = workloadlock_state
            return info

        def is_expected_state(name):
            workloadlock = self.cluster.get_nodeworkloadlock(name)
            if not workloadlock:
                raise ValueError(f"Node workloadlock with name {name} not "
                                 f"found")
            workloadlock_status = workloadlock.data.get('status') or {}
            workloadlock_state = workloadlock_status.get('state')
            return workloadlock_state == state

        waiters.wait(lambda: is_expected_state(name), timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting nodeworkloadlock "
                                 f"state after {timeout} sec.",
                                 status_msg_function=status_message)

    def wait_nodemaintenancerequest_created(self, timeout=900, interval=30):
        """
        Wait while NodeMaintenanceRequest object created
        """
        waiters.wait(lambda: self.cluster.get_nodemaintenancerequests(),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting nodemaintenancerequest for cluster "
                                 f"after {timeout} sec.")

    def wait_nodemaintenancerequest_not_existed(self, timeout=2400, interval=30):
        """
        Wait while NodeMaintenanceRequest object is removed
        """
        waiters.wait(lambda: not self.cluster.get_nodemaintenancerequests(),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting nodemaintenancerequest removing for cluster "
                                 f"after {timeout} sec.")

    def get_clusterworkloadlock_state(self, name):
        workloadlock = self.cluster.get_clusterworkloadlock(name)
        if not workloadlock:
            raise ValueError(f"Cluster workloadlock with name {name} not "
                             f"found")
        workloadlock_status = workloadlock.data.get('status') or {}
        workloadlock_state = workloadlock_status.get('state')
        return workloadlock_state

    def wait_clusterworkloadlock_state(self, name, state='active',
                                       timeout=2400, interval=60):

        def status_message():
            info = dict()
            info['expected_state'] = state
            workloadlock = self.cluster.get_clusterworkloadlock(name)
            workloadlock_status = workloadlock.data.get('status') or {}
            workloadlock_state = workloadlock_status.get('state')
            info['current_state'] = workloadlock_state
            return info

        def is_expected_state(name):
            workloadlock = self.cluster.get_clusterworkloadlock(name)
            if not workloadlock:
                raise ValueError(f"Cluster workloadlock with name {name} not "
                                 f"found")
            workloadlock_status = workloadlock.data.get('status') or {}
            workloadlock_state = workloadlock_status.get('state')
            return workloadlock_state == state

        waiters.wait(lambda: is_expected_state(name), timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting clusterworkloadlock "
                                 f"state after {timeout} sec.",
                                 status_msg_function=status_message)

    def wait_clustermaintenancerequest_created(self, timeout=900, interval=30):
        """
        Wait Cluster ClusterMaintenanceRequest object
        """
        waiters.wait(lambda: self.cluster.get_clustermaintenancerequests(),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting clustermaintenancerequest for cluster "
                                 f"after {timeout} sec.")

    def wait_clustermaintenancerequest_not_existed(self, timeout=900, interval=30):
        """
        Wait Cluster ClusterMaintenanceRequest object
        """
        waiters.wait(lambda: not self.cluster.get_clustermaintenancerequests(),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting clustermaintenancerequest for cluster "
                                 f"after {timeout} sec.")

    def _check_maintenance_status(self, obj, expected_status: bool):
        """
        Chck maintenance status
        Args:
            obj: Cluster or Machine object
            expected_status: bool: expected status

        Returns: bool: is expected status equal to actual status

        """
        obj_status = obj.data.get('status') or {}
        actual_status = obj_status.get('providerStatus', {}).get('maintenance', False)
        LOG.info(f"Expected maintenance status is {expected_status}, actual status is {actual_status}")
        return actual_status == expected_status

    def wait_cluster_maintenance_status(self, expected_status=False, timeout=1200, interval=30):
        """
        Wait expected Cluster Maintenance Request status
        Args:
            expected_status: boolsean value
            timeout: timeout to wait
            interval: time between checks
        Returns: None
        """

        waiters.wait(lambda: self._check_maintenance_status(self.cluster, expected_status),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting expected Cluster Maintenance Request status "
                                 f"after {timeout} sec.")

    def wait_machine_maintenance_status(self, machine_name, expected_status=False, timeout=3600, interval=30):
        """
        Wait expected Machine Maintenance Request status
        Args:
            machine_name: name of expected machine
            expected_status: boolean value
            timeout: timeout to wait
            interval: time between checks
        Returns: None
        """

        machine = self.cluster.get_machine(machine_name)
        waiters.wait(lambda: self._check_maintenance_status(machine, expected_status),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting expected Machine Maintenance Request status "
                                 f"after {timeout} sec.")

    def wait_nodedeletionrequest_created(self, name, timeout=900, interval=30):
        """
        Wait while NodeDeletionRequest object created
        """
        waiters.wait(lambda: self.cluster.get_nodedeletionrequest(name),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting nodedeletionrequest {name} for cluster "
                                 f"after {timeout} sec.")

    def wait_nodedeletionrequest_not_existed(self, name, timeout=900, interval=30):
        """
        Wait while NodeDeletionRequest object is removed
        """
        waiters.wait(lambda: not self.cluster.get_nodedeletionrequest(name),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting nodedeletionrequest {name} removing for cluster "
                                 f"after {timeout} sec.")

    def wait_machine_prepare_deletion_phase(self, machine_name, expected_phase='completed', timeout=1800, interval=30):
        """
        Wait expected Machine Prepare Deletion Phase
        Args:
            machine_name: name of expected Machine
            expected_phase: ('started', 'completed')
            timeout: timeout to wait
            interval: time between checks
        Returns: None
        """

        def get_phase():
            machine = self.cluster.get_machine(machine_name)
            if machine:
                try:
                    actual_phase = machine.get_prepare_deletion_phase()
                except ApiException as e:
                    LOG.error(f"Couldn't get status.providerStatus.prepareDeletionPhase "
                              f"for Machine '{machine_name}':\n{e}")
                    return False
            else:
                if expected_phase in ('completed', ''):
                    actual_phase = expected_phase
                else:
                    raise Exception(f'machine {machine_name} not found in cluster {self.cluster.name}')
            LOG.info(f"Expected machine prepare deletion phase is {expected_phase}, actual is {actual_phase}")
            return expected_phase == actual_phase

        waiters.wait(lambda: get_phase(),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting expected LCMMachine prepare deletion phase"
                                 f"after {timeout} sec.")

    def _check_graceful_reboot_request(self, expected_status: bool):
        exists = self.cluster.get_gracefulrebootrequest() is not None
        return exists == expected_status

    def wait_graceful_reboot_request(self, expected_status=True, timeout=300, interval=30):
        """Wait expected Graceful Reboot Request status
        Args:
            expected_status: True if expecting existence of the object, False if expecting absence
            timeout: timeout to wait
            interval: time between checks
        Returns: None
        """
        waiters.wait(lambda: self._check_graceful_reboot_request(expected_status),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Timeout for waiting expected Graceful RebootRequest status "
                                 f"after {timeout} sec.")

    def _check_machines_reboot(self, boot_time_dict, reboot_state: dict):
        boot_time_dict_new = self.cluster.get_boot_time_dict(exclude_bastion=True)
        not_rebooted = []
        for m_name, m_rebooted in reboot_state.items():
            if m_rebooted:
                continue
            boot_time_old = dateutil.parser.isoparse(boot_time_dict[m_name])
            boot_time_new = dateutil.parser.isoparse(boot_time_dict_new[m_name])
            if boot_time_new > boot_time_old:
                reboot_state[m_name] = True
                continue
            not_rebooted.append(m_name)
        if len(not_rebooted) == 0:
            LOG.info('All requested machines are rebooted')
            return True
        LOG.info(f"Following machines are not rebooted yet: {', '.join(not_rebooted)}")
        return False

    def wait_machines_reboot(self, machines_boot_time_dict, timeout=1200, interval=60):
        """Wait for machines to be rebooted
        Args:
            machines_boot_time_dict: dict of LCM Machines: name -> boot time. Must be populated before
                                     requesting a reboot. Machines from the dict will be waited for reboot
                                     by polling for a newer boot time.
            timeout: timeout to wait
            interval: time between checks
        Returns: None
        """
        rebooted = {name: False for name in machines_boot_time_dict.keys()}
        waiters.wait(lambda: self._check_machines_reboot(machines_boot_time_dict, rebooted),
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=f"Machine reboot timeout after {timeout} sec.")

    def get_ceph_status_details(self) -> dict:
        """
        Get Ceph status directly from the ceph tools pod

        Returns: dict of status details

        """
        cmd_status = ['/bin/sh', '-c', 'ceph -s -f json']
        ceph_tools_pod = self.cluster.get_ceph_tool_pod()
        try:
            output = ceph_tools_pod.exec(cmd_status)
        except Exception:
            LOG.error("Failed to execute command {}". format(" ".join(cmd_status)))
            return {}
        try:
            parsed_output = yaml.safe_load(output)
            if not isinstance(parsed_output, dict):
                LOG.error(f"Cannot convert command output to dictionary: {parsed_output}")
        except yaml.YAMLError:
            LOG.error(f"Failed to decode the output: {output}")
            return {}
        return parsed_output

    def get_ceph_health_status(self):
        """
        Get Ceph Cluster health from kaascephcluster object
        """
        ceph_data = self.cluster.get_kaascephcluster().data
        ceph_data_status = ceph_data.get('status') or {}
        ceph_health = ceph_data_status.get(
            'fullClusterInfo', {}).get('clusterStatus', {}).get('ceph', {}).get('health')
        return ceph_health

    def get_ceph_phase(self):
        """
        Get Ceph Cluster reconcile phase from kaascephcluster object
        """
        ceph_data = self.cluster.get_kaascephcluster().data
        ceph_data_status = ceph_data.get('status') or {}
        ceph_phase = ceph_data_status.get(
            'fullClusterInfo', {}).get('clusterStatus', {}).get('phase')
        return ceph_phase

    def get_ceph_state(self):
        """
        Get Ceph Cluster common status from kaascephcluster object
        """
        ceph_data = self.cluster.get_kaascephcluster().data
        ceph_data_status = ceph_data.get('status') or {}
        ceph_state = ceph_data_status.get(
            'fullClusterInfo', {}).get('clusterStatus', {}).get('state')
        return ceph_state

    def wait_ceph_health_status(self, expected_status='HEALTH_OK', timeout=600,
                                interval=30):
        waiters.wait(
            lambda: self.get_ceph_health_status() == expected_status,
            timeout=timeout, interval=interval,
            timeout_msg=f"Ceph status failed to became "
                        f"{expected_status} in {timeout} sec")

    def get_kaascephcluster_state(self, name=None):
        ceph_cluster_data = self.cluster.get_kaascephcluster(name).data
        ceph_cluster_data_status = ceph_cluster_data.get('status') or {}
        ceph_status = ceph_cluster_data_status.get('shortClusterInfo', {}).get('state')
        return ceph_status

    def wait_kaascephcluster_state(self, expected_state='Ready',
                                   timeout=1800, interval=60):
        waiters.wait(
            lambda: self.get_kaascephcluster_state() == expected_state,
            timeout=timeout,
            interval=interval,
            timeout_msg=f"Timeout waiting for kaascephcluster become "
                        f"{expected_state} state in {timeout} sec. Current "
                        f"state: "
                        f"{self.get_kaascephcluster_state()}")

    def is_default_storage_class_exists(self, name):
        try:
            storage_class = self.cluster.k8sclient.storageclass.read(name)
            LOG.info(f"Cluster default storage class {storage_class}")
            return True
        except Exception as e:
            LOG.debug(e)
            return False

    def wait_default_storage(self, ceph_cluster_pools, timeout=1800, interval=60):
        name = [f"{pool.get('name')}-{pool.get('deviceClass')}" for
                pool in ceph_cluster_pools if pool.get('default')]
        if not name:
            LOG.warning('Default Storage class not found')
            return
        waiters.wait(
            lambda: self.is_default_storage_class_exists(name[0]) is True,
            timeout=timeout,
            interval=interval,
            timeout_msg=f"Timeout waiting for default storage: {name}"
        )

    def get_cephoperationrequest_state(self, name, namespace, operation_key=None):
        ceph_request = self.cluster.k8sclient.kaas_cephoperationrequests.get(name=name, namespace=namespace)
        ceph_request_data = ceph_request.data
        ceph_request_status = ceph_request_data.get('status') or {}
        if not ceph_request_status:
            LOG.debug(f"Request object {name} is not ready yet. Object data:\n{yaml.dump(ceph_request_data)}")
            return None, None
        else:
            request_phase = ceph_request_status.get('phase', None)
            operation_phase = ceph_request_status.get(operation_key, {}).get('phase', None) if operation_key else None
            return request_phase, operation_phase

    def wait_cephoperationrequest_state(self, name, namespace, expected_request_phase='Completed',
                                        expected_operation_phase=None, operation_key=None,
                                        timeout=1200, interval=120):

        def status_message_func():
            overall_status = self.get_cephoperationrequest_state(name, namespace, operation_key=operation_key)
            request_phase = overall_status[0]
            operation_phase = overall_status[1]
            msg = f"Current request phase is: {request_phase}. Expected: {expected_request_phase}. "
            if operation_key:
                msg += (f"Current operation {operation_key} phase is: {operation_phase}. "
                        f"Expected: {expected_operation_phase}")
            return msg
        waiters.wait(
            lambda: self.get_cephoperationrequest_state(
                name=name, namespace=namespace, operation_key=operation_key) == (expected_request_phase,
                                                                                 expected_operation_phase),
            timeout=timeout,
            interval=interval,
            status_msg_function=status_message_func)

    def get_ceph_daemons_status(self):
        ceph_data = self.cluster.get_kaascephcluster().data
        ceph_data_status = ceph_data.get('status') or {}
        ceph_daemons_status = ceph_data_status.get('fullClusterInfo', {}).get('daemonsStatus')
        return ceph_daemons_status

    def storageclass_exists(self, sc_prefix='kubernetes-'):
        storage_classes = \
            self.cluster.k8sclient.api_storage.list_storage_class().to_dict()[
                'items']
        return len([x for x in storage_classes if
                    x['metadata']['name'].startswith(sc_prefix)]) > 0

    def get_ceph_pvc_status(self, pvc_name, namespace):
        try:
            pvc = self.cluster.k8sclient.pvolumeclaims.get(name=pvc_name, namespace=namespace)
            pvc_status = pvc.data.get('status') or {}
            phase = pvc_status.get('phase', '')
            return phase
        except ApiException as e:
            LOG.error(f"Couldn't get pvc status for pvc {pvc_name} in namespace {namespace}."
                      f"Reason: {e} ")
            return None

    def wait_ceph_pvc_status(self, pvc_name, namespace, expected_status='Bound', timeout=600, interval=30):
        """
        Available statuses:
        'Bound', 'Lost', 'Pending'
        https://kubernetes.io/docs/reference/kubernetes-api/config-and-storage-resources/persistent-volume-claim-v1/#PersistentVolumeClaimStatus # noqa
        """
        msg = (f"Timeout waiting for pvc {pvc_name} status to be {expected_status}. "
               f"Current status: {self.get_ceph_pvc_status(pvc_name=pvc_name, namespace=namespace)}")
        waiters.wait(lambda: self.get_ceph_pvc_status(pvc_name=pvc_name, namespace=namespace) == expected_status,
                     timeout=timeout, interval=interval, timeout_msg=msg)

    def check_ceph_pvc(self,
                       base_image_repo=None,
                       cephfs_enabled=False,
                       sc_name='kubernetes-',
                       device_class=None):
        """
        Simple test for ceph pvc in child-cluster
        Scenario:
            1. Spawn pod with attached pv on each node
            2. Write some data to file in pv
            3. Detach and delete pod
            4. Spawn new pod and attach to old pv
            5. Check data content
            6. Resize pvc
            7. Delete pod; delete pvc; Delete ns
        """
        base_image_repo = base_image_repo or self.cluster.determine_mcp_docker_registry()
        mount_path = '/data/test'
        pvc_size = 1
        ns = f'test-ns-{utils.gen_random_string(6)}'

        machines = self.cluster.get_machines(machine_type='worker')

        if cephfs_enabled:
            LOG.info("Ceph filesystem: ENABLED")
        LOG.info("Wait for Ceph StorageClass created")
        waiters.wait(lambda: self.storageclass_exists(sc_name),
                     timeout=10 * 60, interval=60)
        storage_classes = \
            self.cluster.k8sclient.api_storage.list_storage_class().to_dict()[
                'items']
        # Search storage class based on name
        storage_class = next(x for x in storage_classes if
                             x['metadata']['name'].startswith(sc_name))
        storage_class_name = storage_class['metadata']['name']
        if not device_class:
            storage_class_pool = storage_class['parameters']['pool']
            device_class = storage_class_pool.split('-')[1]
        LOG.info(f"Found Ceph filesystem's StorageClass {storage_class_name} "
                 f"with {device_class} device class")

        LOG.info('Check Ceph pvc')
        LOG.info(f'Found worker machines: {[o.name for o in machines]}')
        LOG.info(f'Create "{ns}" namespace')
        namespace = self.cluster.k8sclient.namespaces.create(
            name=ns,
            body={'metadata': {'name': ns}})

        pvc_names = []
        if cephfs_enabled:
            pvc_name = f"test-pv-claim-{storage_class_name}"
            pvc_template = templates.render_template(
                settings.CEPH_PVC_YAML, {
                    'PVC_NAME': pvc_name,
                    'PVC_SIZE': f'{pvc_size}Gi',
                    'ACCESS_MODE': 'ReadWriteMany',
                    'STORAGE_CLASS_NAME': storage_class_name})
            pvc_json_body = json.dumps(
                yaml.load(pvc_template, Loader=yaml.SafeLoader))

            LOG.info(f'Creating pvc {pvc_name}')
            self.cluster.k8sclient.pvolumeclaims.create(name=pvc_name,
                                                        namespace=ns,
                                                        body=json.loads(
                                                            pvc_json_body))
            pvc_names.append(pvc_name)
        else:
            for n in machines:
                pvc_name = f"test-pv-claim-{n.uid}"
                pvc_template = templates.render_template(
                    settings.CEPH_PVC_YAML, {
                        'PVC_NAME': pvc_name,
                        'PVC_SIZE': f'{pvc_size}Gi',
                        'ACCESS_MODE': 'ReadWriteOnce',
                        'STORAGE_CLASS_NAME': storage_class_name})
                pvc_json_body = json.dumps(
                    yaml.load(pvc_template, Loader=yaml.SafeLoader))

                LOG.info('Creating pvc')
                self.cluster.k8sclient.pvolumeclaims.create(name=pvc_name,
                                                            namespace=ns,
                                                            body=json.loads(
                                                                pvc_json_body))
                self.wait_ceph_pvc_status(pvc_name=pvc_name, namespace=ns,
                                          expected_status='Bound', timeout=900)
                pvc_names.append(pvc_name)

        try:
            pods = []
            for n in machines:
                if cephfs_enabled:
                    pvc_name = f"test-pv-claim-{storage_class_name}"
                else:
                    pvc_name = f"test-pv-claim-{n.uid}"
                node_labels = n.get_k8s_node().read().metadata.labels
                node_name = node_labels['kubernetes.io/hostname']
                pod_template = templates.render_template(
                    settings.CEPH_PV_CHECK_POD_YAML,
                    {'POD_NAME': f"test-pod-{n.uid}",
                     'PVC_NAME': pvc_name,
                     'MOUNT_PATH': mount_path,
                     'NODE_LABEL': node_name,
                     'CEPH_NGINX_IMAGE_BASE_REPO': base_image_repo})
                pod_json_body = json.dumps(
                    yaml.load(pod_template, Loader=yaml.SafeLoader))

                LOG.info(f'Create pod on {node_name} node ({n.name} machine)')
                pod = self.cluster.k8sclient.pods.create(name=f"test-pod-{n.uid}",
                                                         namespace=ns,
                                                         body=json.loads(
                                                             pod_json_body))
                pods.append(pod)

            for pod in pods:
                LOG.info(f'Waiting for pod {pod.name} is Running.')
                pod.wait_phase(phases='Running', timeout=30*60, interval=30)

            file_paths = {}
            expected_strings = {}
            for pod in pods:
                LOG.info(f"Create new file in attached pv of {pod.name} pod")
                file_path = f'{mount_path}/{utils.gen_random_string(4)}_test.txt'
                expected_string = utils.gen_random_string(10)
                pod.exec(
                    ['/bin/sh', '-c',
                     f'echo "{expected_string}" > {file_path}'])
                pod.check_file_exists(file_path)
                file_paths[pod.name] = file_path
                expected_strings[pod.name] = expected_string

            for pod in pods:
                LOG.info(f"Delete {pod.name} pod")
                pod.delete()

            pods = []
            for n in machines:
                if cephfs_enabled:
                    pvc_name = f"test-pv-claim-{storage_class_name}"
                else:
                    pvc_name = f"test-pv-claim-{n.uid}"
                node_labels = n.get_k8s_node().read().metadata.labels
                node_name = node_labels['kubernetes.io/hostname']
                pod_template = templates.render_template(
                    settings.CEPH_PV_CHECK_POD_YAML,
                    {'POD_NAME': f"test-pod-{n.uid}",
                     'PVC_NAME': pvc_name,
                     'MOUNT_PATH': mount_path,
                     'NODE_LABEL': node_name,
                     'CEPH_NGINX_IMAGE_BASE_REPO': base_image_repo})
                pod_json_body = json.dumps(
                    yaml.load(pod_template, Loader=yaml.SafeLoader))

                LOG.info(f'Re-create pod on {node_name} node ({n.name} machine)')
                pod = self.cluster.k8sclient.pods.create(name=f"test-pod-{n.uid}",
                                                         namespace=ns,
                                                         body=json.loads(
                                                             pod_json_body))
                pods.append(pod)

            for pod in pods:
                LOG.info(f"Waiting for pod {pod.name} is Running.")
                pod.wait_phase(phases='Running', timeout=30 * 60, interval=30)

            for pod in pods:
                LOG.info('Check file content')
                file_path = file_paths[pod.name]
                expected_string = expected_strings[pod.name]
                pod.check_file_exists(file_path)
                res = pod.exec(['/bin/sh', '-c', f'cat {file_path}'])
                assert res.rstrip() == expected_string, \
                    (f"Expected file in pod {pod.name} has unexpected content: "
                     f"expected: '{expected_string}', actual: {res.rstrip()}")

            # If baremetal and has special parameter https://mirantis.jira.com/browse/PRODX-16779
            if self.cluster.provider == utils.Provider.baremetal:
                storage_name = storage_class_name.split('-')[0]
                ceph_crd = self.cluster.get_kaascephcluster()
                ceph_pool = next((p for p in ceph_crd.data['spec']['cephClusterSpec']['pools'] if
                                  p['name'] == storage_name and p['deviceClass'] == device_class), {})

                pod_pvc_map = {}
                pvc_pod_map = {}
                for n in machines:
                    pvc_name = f"test-pv-claim-{n.uid}"
                    if cephfs_enabled:
                        pvc_name = f"test-pv-claim-{storage_class_name}"
                    pod_pvc_map[f"test-pod-{n.uid}"] = pvc_name
                    pvc_pod_map[pvc_name] = f"test-pod-{n.uid}"

                if cephfs_enabled or ceph_pool.get('allowVolumeExpansion', False):
                    pvcs = {}
                    pvc_incr = pvc_size
                    if cephfs_enabled:
                        pvc_incr += 1
                        LOG.info(f'Resize pv to {pvc_incr} Gi')
                        pvc_template = templates.render_template(
                            settings.CEPH_PVC_YAML, {
                                'PVC_NAME': f"test-pv-claim-{storage_class_name}",
                                'PVC_SIZE': f'{pvc_incr}Gi',
                                'ACCESS_MODE': 'ReadWriteMany',
                                'STORAGE_CLASS_NAME': storage_class_name})
                        pvc_json_body = json.dumps(
                            yaml.load(pvc_template, Loader=yaml.SafeLoader))
                        pvc = self.cluster.k8sclient.pvolumeclaims.update(name=f"test-pv-claim-{storage_class_name}",
                                                                          namespace=ns,
                                                                          body=json.loads(
                                                                              pvc_json_body))
                        pvcs[pvc_incr] = pvc
                    else:
                        for pod in pods:
                            pvc_incr += 1
                            LOG.info(f'Resize pv to {pvc_incr} Gi')
                            pvc_template = templates.render_template(
                                settings.CEPH_PVC_YAML, {
                                    'PVC_NAME': pod_pvc_map[pod.name],
                                    'PVC_SIZE': f'{pvc_incr}Gi',
                                    'ACCESS_MODE': 'ReadWriteOnce',
                                    'STORAGE_CLASS_NAME': storage_class_name})
                            pvc_json_body = json.dumps(
                                yaml.load(pvc_template, Loader=yaml.SafeLoader))
                            pvc = self.cluster.k8sclient.pvolumeclaims.update(name=pod_pvc_map[pod.name],
                                                                              namespace=ns,
                                                                              body=json.loads(
                                                                                  pvc_json_body))
                            pvcs[pvc_incr] = pvc

                    for pod in pods:
                        LOG.info(f"Waiting for pod {pod.name} is Running.")
                        pod.wait_phase(phases='Running')

                    for pvc_incr, pvc in pvcs.items():
                        pvc_volume_name = pvc.data['spec']['volume_name']

                        LOG.info('Check new pv size')
                        pv = next(x for x in self.cluster.k8sclient.pvolumes.list_all() if
                                  x.data['spec']['claim_ref'] is not None
                                  and x.name == pvc_volume_name)
                        waiters.wait(
                            lambda: pv.data['spec']['capacity']['storage'] == f'{pvc_incr}Gi',
                            timeout=30, interval=5, timeout_msg=f"Expected pvc_size = {pvc_incr}Gi in pod "
                                                                f"{pvc_pod_map[pvc.name]} but we have "
                                                                f"{pv.data['spec']['capacity']['storage']}")
                else:
                    LOG.info(f'allowVolumeExpansion parameter not found or in FALSE state.'
                             f'Skip pvc resize test. Print ceph_pool: {ceph_pool}')
            else:
                LOG.info('Skip pvc resize test for non-BM provider')

            for pod in pods:
                LOG.info(f"Delete {pod.name} pod")
                pod.delete()

        finally:
            LOG.info('Delete pvcs')
            for pvc_name in pvc_names:
                pvc = self.cluster.k8sclient.pvolumeclaims.get(name=pvc_name, namespace=ns)
                pvc.delete()

            LOG.info('Delete namespace')
            namespace.delete()
            LOG.info('Check Ceph pvc done')

    def check_k8s_nodes(self, timeout=360, interval=10):
        LOG.info("Checking k8s nodes status")
        try:
            waiters.wait(
                lambda: self.k8sclient.nodes.all_ready(),
                timeout=timeout, interval=interval)
        except exceptions.TimeoutError:
            nodes = {key: value for key, value in
                     self.k8sclient.nodes.list_statuses().items()
                     if value != 'True'}
            err = f"Timeout waiting for nodes to be Ready. " \
                  f"After {timeout} sec next nodes are not Ready: {nodes}"
            raise exceptions.TimeoutError(err)
        LOG.info("All k8s nodes are Ready")

    def wait_k8s_node_status(self, nodename='', expected_status='',
                             timeout=360, interval=10):
        LOG.info(f"Wait for k8s node stay in {expected_status} status.")
        try:
            waiters.wait(
                lambda: self.k8sclient.nodes.node_status(nodename,
                                                         expected_status),
                timeout=timeout, interval=interval,
                timeout_msg=f"Node not in status {expected_status}")
        except exceptions.TimeoutError:
            err = f"Timeout waiting for node {nodename} to be "\
                  f"{expected_status} After {timeout} sec."
            raise exceptions.TimeoutError(err)
        LOG.info(f"k8s node {nodename} is in status <{expected_status}>.")

    # note: check_machines_status will do similar thing, but we still need this
    def check_cluster_nodes(self, timeout=1800, interval=10,
                            expected_state='Ready'):
        LOG.info("Checking cluster nodes status")
        try:
            waiters.wait(lambda: self.cluster.cluster_status.lower() ==
                         expected_state.lower(),
                         timeout=timeout, interval=interval)
            LOG.info(f"All cluster nodes are {expected_state}")
        except exceptions.TimeoutError:
            status = self.cluster.data['status']
            if not status:
                raise Exception("Cluster is in Pending status")
            nodes = status.get('providerStatus', {}).get('nodes', {})
            if not nodes:
                raise Exception(f"Nodes for cluster are not found {status}")
            err = f"Timeout waiting for cluster " \
                  f"nodes to be {expected_state} " \
                  f"after {timeout} sec: {nodes}"
            raise exceptions.TimeoutError(err)

    def check_pods_number(self, pod_name, ns, pods):
        """Wait till replica number restores"""
        def correct_number_of_pods(pods):
            # let's take number of pods we had
            number = len(pods)
            # extract old names (we expect to see the same pods)
            pods_names = [x.name for x in pods]
            new_list = self.k8sclient.pods.list(
                ns, field_selector="status.phase=Running"
            )
            # check if new list has same number of pods (with same names)
            new_pods_len = len([x for x in new_list if x.name in pods_names])
            return True if number == new_pods_len else False
        try:
            waiters.wait(
                lambda: correct_number_of_pods(pods),
                timeout=360, interval=10)
        except exceptions.TimeoutError:
            actual_num = len(self.k8sclient.pods.list_starts_with(
                pod_name, ns, field_selector="status.phase=Running"))
            LOG.error(f'Timeout waiting for correct number. '
                      f'Actual number: {actual_num}')
            raise TimeoutError

    def _get_replicasets_by_namespace(self, target_namespaces):
        rs_info = {}
        if isinstance(target_namespaces, str):
            target_namespaces = [target_namespaces]
        all_rs = self.k8sclient.replicasets.list_all()
        for rs in all_rs:
            if target_namespaces and rs.namespace not in target_namespaces:
                continue
            rs.kind = 'ReplicaSet'  # ReplicaSet objects in list have 'kind: None' until rs.read()
            rs.spec = rs._read_cache.spec  # spec in _read_cache from ReplicaSetList contains required values
            rs_info.setdefault(rs.namespace, {})[f"{rs.kind}/{rs.uid}"] = rs
        return rs_info

    def check_k8s_pods(self, phases=('Running', 'Succeeded'),
                       target_namespaces=None,
                       timeout=settings.WAIT_PODS_READY_TIMEOUT,
                       interval=30, pods_prefix=''):
        """Wait till all expected pods for cluster are in specified
           phase and have Ready=True for all containers
        Args:
            phases: list of expected pod phases
            target_namespaces: namespace (str) or namespaces (list)
                               where pods should be checked
            timeout: timeout to wait
            interval: time between checks
        """

        def wait_for_running_pods(pause=60):
            """Get the pods statuses and compare the restart counts

            :param pause: int, seconds, pause between restart checks
            """
            LOG.info("Check k8s pods status")
            rs_info = self._get_replicasets_by_namespace(target_namespaces)
            # 404 appeared when a just created Management cluster don't have /clusters api-resource yet
            # or sometimes resource can disappear at checking procedure, and we will receive 404
            try:
                pods_info = self.k8sclient.pods.check_pods_statuses(
                    target_namespaces=target_namespaces, phases=phases,
                    excluded_pods=self.EXCLUDED_PODS,
                    pods_prefix=pods_prefix, replicasets=rs_info)
                if not pods_info:
                    return False
            except ApiException as ex:
                if ex.status != 404:
                    raise ex
                else:
                    LOG.error(f"Error happened while checking pods statuses: {ex}")
                    return False

            LOG.info(f"Wait for {pause} seconds before the second check")
            time.sleep(pause)

            LOG.info("Check k8s pods status one more time")
            rs_info = self._get_replicasets_by_namespace(target_namespaces)
            try:
                pods_info_new = self.k8sclient.pods.check_pods_statuses(
                    target_namespaces=target_namespaces, phases=phases,
                    excluded_pods=self.EXCLUDED_PODS,
                    pods_prefix=pods_prefix, replicasets=rs_info)
                if not pods_info_new:
                    return False
            except ApiException as ex:
                if ex.status != 404:
                    raise ex
                else:
                    LOG.error(f"Error happened while checking pods statuses: {ex}")
                    return False

            result = True
            for pod, info in pods_info.items():
                if pod not in pods_info_new:
                    LOG.warning(f"Pod {pod} disappeared after {pause} seconds")
                    result = False
                    continue
                cont_restarts_new = pods_info_new[pod]['containers_restarts']
                for container, restarts in info['containers_restarts'].items():
                    if container not in cont_restarts_new:
                        LOG.warning(f"Container {container} from pod {pod} "
                                    f"disappeared after {pause} seconds")
                        result = False
                        continue
                    restarts_new = cont_restarts_new[container]
                    if restarts_new > restarts:
                        LOG.warning(f"Container {container} from pod {pod} "
                                    f"has been restarted after {pause} seconds"
                                    f", current restarts: {restarts_new}")
                        result = False

            if result:
                LOG.info("All pods are in correct state")
            return result

        LOG.info("Checking k8s pods phase and containers status")
        if not target_namespaces:
            target_namespaces = self.cluster.expected_pods.keys()
        try:
            # Wait for pods status
            waiters.wait(wait_for_running_pods,
                         timeout=timeout, interval=interval)

            # Wait for jobs
            waiters.wait(
                lambda: self.k8sclient.jobs.check_jobs_completed(
                    target_namespaces=target_namespaces,
                    excluded_jobs=self.EXCLUDED_JOBS,
                    jobs_prefix=pods_prefix),
                timeout=timeout, interval=interval)
        except exceptions.TimeoutError:
            pods = self.k8sclient.pods.list_raw().to_dict()['items']
            if target_namespaces:
                if isinstance(target_namespaces, str):
                    target_namespaces = [target_namespaces]
                pods = [pod for pod in pods
                        if pod['metadata']['namespace'] in target_namespaces]
            utils.print_pods_status(pods)
            failed_pods = {}
            ready_text = " and Ready=True for " \
                         "each container"
            for pod, info in \
                    self.k8sclient.pods.get_pods_statuses(pods).items():
                # Check for the expected pod phase
                if info['phase'] not in phases:
                    failed_pods[pod] = info
                # Check for containers status inside the non-completed pod
                if (False in info['containers'].values()
                        and info['phase'] not in 'Succeeded'):
                    failed_pods[pod] = info
            err = f"Timeout waiting for pods statuses. " \
                  f"After {timeout} sec next pods are not in {phases} phase" \
                  f"{ready_text}: {failed_pods}"

            raise TimeoutError(err)
        LOG.info("All pods and jobs are in correct state")

    def check_ha_kill_proc_pods(self, pod_group, command):
        """Kill processed in every container in all cluster pods
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every namespace and
        expected pod entry in get_expected_pods
        Note: 'kill -- - 1' will kill all processes except pid 1

        Scenario:
            1. Iterate by each pod in a group
            2. Iterate by each container in a pod and
               kill processes in the container
            3. Wait till number of replicas will be restored
            4. Check pods and container statuses in this group
               (Running and Ready)
            5. Check all pods phases (must be Running or Completed)
               (in fixture)

        Expected result - pods are recreated, number of replicas is restored.
        """
        if settings.RUN_HA_TESTS:
            pod_name, pod_num, pods = pod_group

            ns = [pod.namespace for pod in pods][0]
            LOG.info("Iterate by each pod in a group")
            for pod in pods:
                # let's reread every pod to make sure it is healthy
                # before killing containers
                pod_data = pod.read()
                assert False not in [
                    x.ready for x in pod_data.status.container_statuses], \
                    f"Some containers are not ready for {pod.name} " \
                    f"pod: {pod_data.status.container_statuses}"
                LOG.info(f"Killing services in "
                         f"all containers of the pod: {pod.name}")
                for cont in pod_data.status.container_statuses:
                    restarts = cont.restart_count
                    LOG.info(f"Execute command {command} "
                             f"in {cont.name} container. "
                             f"Restarts: {restarts}")
                    try:
                        result = pod.exec(['/bin/sh', '-c', command],
                                          container=cont.name)
                        if result:
                            LOG.info(f"Killed with result: {result}")
                    except (ApiException, WebSocketBadStatusException) as e:
                        LOG.error(e)
                    continue
            LOG.info("Check number of pods")
            self.check_pods_number(pod_name, ns, pods)
            LOG.info(f"Check status for {pod_name} pods")
            # PRODX-12786
            if 'kibana' in pod_name:
                time.sleep(30)
                self.check_k8s_pods(pods_prefix=pod_name,
                                    timeout=600,
                                    interval=30)
            else:
                self.check_k8s_pods(pods_prefix=pod_name,
                                    timeout=300,
                                    interval=30)
        else:
            LOG.info("\nHA tests for services in pod containers were "
                     "skipped as RUN_HA_TESTS was set to False\n")

    def get_leader(self, app_prefix):
        def get_leader_pod_name(cms):
            configmap_annotation = cms[0].read().to_dict()['metadata']['annotations']
            leader_data = configmap_annotation['control-plane.alpha.kubernetes.io/leader']
            leader_pod_name = json.loads(leader_data).get('holderIdentity')
            return leader_pod_name
        cms = [cm for cm in self.k8sclient.configmaps.list_all(name_prefix=app_prefix + '-leader-election')]
        assert len(cms) == 1, f"Configmap for app {app_prefix} not found or has multiple name based matches"
        waiters.wait(lambda: get_leader_pod_name(cms), timeout=60, interval=5)

        leader_pod_name = get_leader_pod_name(cms)
        LOG.info(f"Current leader for app {app_prefix} is {leader_pod_name}")
        return leader_pod_name

    def check_ha_delete_leader_pod(self, app_group=None):
        """Detect and delete leader pod for selected app. Check that leader changed
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for selected application

        Scenario:
            1. Detect leader
            2. Delete leader pod
            3. Wait till leader was changed

        Expected result - pods are recreated, number of replicas is restored.
        """
        if settings.RUN_HA_TESTS:
            pods = self.cluster.k8sclient.pods.list_all(name_prefix=app_group)
            assert len(pods) > 0, "Pods not found for selected app"
            leader_pod_name_before = self.get_leader(app_prefix=app_group).split("_")[0]

            LOG.info("Get leader pod and delete it")
            for pod in pods:
                if pod.name == leader_pod_name_before:
                    LOG.info(f"Deleting leader pod: {pod.name}")
                    pod.delete(timeout=180)

            def wait_leader_changed(leader_before=None):
                leader_pod_name_after = self.get_leader(app_prefix=app_group).split("_")[0]
                if leader_pod_name_after != leader_before:
                    return True
                else:
                    LOG.info(f"Leader is not changed yet. Should be not {leader_before}")
                    return False

            waiters.wait(lambda: wait_leader_changed(leader_pod_name_before),
                         timeout=120, interval=5)
        else:
            LOG.info("\nHA tests were skipped as RUN_HA_TESTS was set to False\n")

    def check_ha_delete_pods(self, pod_group):
        """Delete all pods in any cluster one by one
         Precondition - all expected pods and their replicas must be presented
         The following scenario is executed for every namespace and
         expected pod entry in get_expected_pods

         Scenario:
             1. Compare actual number of replicas for pod with expected
             2. Iterate by each replica
             3. Delete pod (1 replica)
             4. Wait till number of replicas will be restored
             5. Check pods statuses in this group (Running and Ready)

         Expected result - pods are recreated, number of replicas is restored.
         """
        if settings.RUN_HA_TESTS:
            pod_name, pod_num, pods = pod_group
            ns = [pod.namespace for pod in pods][0]
            LOG.info("Deleting pods")
            for pod in pods:
                LOG.info(f"Deleting pod: {pod.name}")
                pod.delete(timeout=180)
                # special handling of mariadb-server StatefulSets
                if 'mariadb-server' in pod.name:
                    LOG.info(f"Checking readiness "
                             f"for {pod.name} (StatefulSet)")
                    pod.wait_ready()
                LOG.info("Check number of pods")
                self.check_pods_number(pod_name, ns, pods)
                LOG.info(f"Check status for {pod_name} pods")
                self.check_k8s_pods(pods_prefix=pod_name,
                                    timeout=300,
                                    interval=30)
        else:
            LOG.info("\nHA tests for pods were skipped "
                     "as RUN_HA_TESTS was set to False\n")

    def check_ha_kill_containerd_shim(self):
        """Kill all containerd-shim processes on every cluster node
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to node and run kill command for containerd-shim processes
            2. Wait
            3. Check that all pods are Running and Ready

        Expected result - pods and their containers are Running and Ready.
        Number of replicas are the same.
        """
        if settings.RUN_HA_TESTS:
            nodes = self.cluster.get_machines()
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                node._run_cmd("for i in $(ps uax | grep containerd-shim | "
                              "grep -v grep | awk '{print $2}'); "
                              "do sudo kill -9 $i; done",
                              ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                time.sleep(60)
                LOG.info("Waiting for k8s to be ready")
                waiters.wait_pass(
                    lambda: self.k8sclient.pods.list_all(), timeout=120)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30)
                # TODO: maybe check pod restarts also?
                # additonal timeout ?
        else:
            LOG.info("\nHA tests to kill containerd-shim were skipped "
                     "as RUN_HA_TESTS was set to False\n")

    def check_ha_kill_mariadb_cluster(self):
        """Kill mysqld service on every cluster node (one at a time)

        Executed only on MOSK cluster

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to node
            2. Kill mysqld service
            3. Check mariadb become ready
            4. Check that all pods are Running and Ready

        Expected result - all mysqld services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version:
            nodes = self.cluster.get_machines(k8s_labels={"openstack-control-plane": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_mariadb_mariadb-server-.?_openstack'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                mysqld_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | grep mysqld| cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {mysqld_pid}",
                              ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill mariadb were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK\n")
            pytest.skip("HA tests to kill mariadb were skipped")

    def check_ha_kill_memcached_cluster(self):
        """Kill memcached service on every cluster node (one at a time)

        Executed only on MOSK cluster

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to node
            2. Kill memcached service
            3. Check memcached become ready
            4. Check that all pods are Running and Ready

        Expected result - all memcached services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version:
            nodes = self.cluster.get_machines(k8s_labels={"openstack-control-plane": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | "
                    "grep -E '^k8s_memcached_openstack-memcached-memcached-.?_openstack'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                memcached_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | grep memcached | cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {memcached_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces='openstack')
        else:
            LOG.info("\nHA tests to kill memcached were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK\n")
            pytest.skip("HA tests to kill memcached were skipped")

    def check_ha_kill_etcd_cluster(self):
        """Kill etcd service on every cluster node (one at a time)

        Executed only on MOSK cluster

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to node
            2. Kill etcd service
            3. Check etcd become ready
            4. Check that all pods are Running and Ready

        Expected result - all etcd services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version:
            nodes = self.cluster.get_machines(k8s_labels={"openstack-control-plane": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_etcd_etcd-etcd-.?_openstack'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                etcd_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | grep etcd | cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {etcd_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill etcd were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK\n")
            pytest.skip("HA tests to kill etcd were skipped")

    def check_ha_kill_rabbitmq(self):
        """Kill rabbtimq service

        Executed only on MOSK cluster

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found node where rabbitmq is running
            2. SSH to node
            3. Kill beam.smp service
            4. Check rabbtimq become ready
            5. Check that all pods are Running and Ready

        Expected result - rabbtimq services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version:
            pod_name = "openstack-rabbitmq-rabbitmq-0"
            rq_pod = self.cluster.k8sclient.pods.get(pod_name, namespace="openstack")
            node = self.cluster.get_machine_by_k8s_name(rq_pod.node_name)
            cont_name = node._run_cmd(
                "docker ps --format '{{.Names}}' | grep -E '^k8s_rabbitmq_openstack-rabbitmq-rabbitmq-0_openstack'",
                ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
            beam_smp_pid = node._run_cmd(
                f"docker container top {cont_name} -o pid,args | grep beam.smp | cut -d ' ' -f 1",
                ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
            node._run_cmd(f"sudo kill -9 {beam_smp_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
            LOG.info("Waiting for rabbitmq to be in a correct state")
            self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill rabbitmq were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK\n")
            pytest.skip("HA tests to kill rabbitmq were skipped")

    def check_ha_kill_neutron_rabbitmq(self):
        """Kill neutron rabbtimq service

        Executed only on MOSK cluster

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found node where rabbitmq is running
            2. SSH to node
            3. Kill beam.smp service
            4. Check rabbtimq become ready
            5. Check that all pods are Running and Ready

        Expected result - rabbtimq services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version:
            pod_name = "openstack-neutron-rabbitmq-rabbitmq-0"
            rq_pod = self.cluster.k8sclient.pods.get(pod_name, namespace="openstack")
            node = self.cluster.get_machine_by_k8s_name(rq_pod.node_name)
            cont_name = node._run_cmd(
                "docker ps --format '{{.Names}}' | "
                "grep -E '^k8s_rabbitmq_openstack-neutron-rabbitmq-rabbitmq-0_openstack'",
                ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
            beam_smp_pid = node._run_cmd(
                f"docker container top {cont_name} -o pid,args | grep beam.smp | cut -d ' ' -f 1",
                ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
            node._run_cmd(f"sudo kill -9 {beam_smp_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
            LOG.info("Waiting for rabbitmq to be in a correct state")
            self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill neutron rabbtimq were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK\n")
            pytest.skip("HA tests to kill neutron rabbtimq were skipped")

    def check_ha_kill_neutron_ovs_agent(self):
        """Kill neutron ovs agent

        Executed only on MOSK cluster

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found nodes where neutron ovs agent is running
            2. SSH to node
            3. Kill neutron-openvswitch-agent service
            4. Check neutron ovs agent become ready
            5. Check that all pods are Running and Ready

        Expected result - rabbtimq services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version:
            nodes = self.cluster.get_machines(
                k8s_labels={"openvswitch": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_neutron-ovs-agent_neutron-ovs-agent'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                ovs_agent_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | grep neutron-openvswitch-agent | cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {ovs_agent_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill neutron-openvswitch-agent were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK\n")
            pytest.skip("HA tests to kill neutron-openvswitch-agent were skipped")

    def check_ha_kill_neutron_dhcp_agent(self):
        """Kill neutron dhcp agent

        Executed only on MOSK cluster

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found nodes where neutron dhcp agent is running
            2. SSH to node
            3. Kill neutron-dhcp-agent service
            4. Check dhcp agent become ready
            5. Check that all pods are Running and Ready

        Expected result - neutron-dhcp-agent services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version:
            nodes = self.cluster.get_machines(
                k8s_labels={"openstack-gateway": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_neutron-dhcp-agent_neutron-dhcp-agent'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                dhcp_agent_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | grep neutron-dhcp-agent | cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {dhcp_agent_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill neutron-dhcp-agent were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK\n")
            pytest.skip("HA tests to kill neutron-dhcp-agent were skipped")

    def check_ha_kill_neutron_l3_agent(self):
        """Kill neutron l3 agent

        Executed only on MOSK cluster

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found nodes where neutron l3 agent is running
            2. SSH to node
            3. Kill neutron-l3-agent service
            4. Check l3 agent become ready
            5. Check that all pods are Running and Ready

        Expected result - l3 agent services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version:
            nodes = self.cluster.get_machines(
                k8s_labels={"openstack-gateway": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_neutron-l3-agent_neutron-l3-agent'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                l3_agent_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | grep neutron-l3-agent | cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {l3_agent_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill neutron l3 agent were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK\n")
            pytest.skip("HA tests to kill neutron l3 agent were skipped")

    def check_ha_kill_tf_control(self):
        """Kill contrail-control

        Executed only on MOSK cluster with TF

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found nodes where contrail-control is running
            2. SSH to node
            3. Kill contrail-control service
            4. Check l3 agent become ready
            5. Check that all pods are Running and Ready

        Expected result - contrail-control services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version and self.cluster.tf_enabled():
            nodes = self.cluster.get_machines(
                k8s_labels={"tfcontrol": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_control_tf-control'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                contrail_control_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | grep contrail-control | cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {contrail_control_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="tf")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill tf control were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK with TF\n")
            pytest.skip("HA tests to kill tf control were skipped")

    def check_ha_kill_tf_kafka(self):
        """Kill tf kafka

        Executed only on MOSK cluster with TF

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found nodes where tf kafka is running
            2. SSH to node
            3. Kill tf kafka service
            4. Check l3 agent become ready
            5. Check that all pods are Running and Ready

        Expected result - tf kafka services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version and self.cluster.tf_enabled():
            nodes = self.cluster.get_machines(
                k8s_labels={"tfanalyticsdb": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_kafka-broker_tf-kafka'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                kafka_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | "
                    "grep /etc/kafka/kafka.properties | cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {kafka_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="tf")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill tf kafka were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK with TF\n")
            pytest.skip("HA tests to kill tf kafka were skipped ")

    def check_ha_kill_tf_zookeeper(self):
        """Kill tf zookeeper

        Executed only on MOSK cluster with TF

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found nodes where tf zookeeper is running
            2. SSH to node
            3. Kill tf zookeeper service
            4. Check tf zookeeper agent become ready
            5. Check that all pods are Running and Ready

        Expected result - tf zookeeper services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version and self.cluster.tf_enabled():
            nodes = self.cluster.get_machines(
                k8s_labels={"tfconfigdb": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_zookeeper_tf-zookeeper-.?_tf'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                zookeeper_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | grep /data/conf/zoo.cfg | cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {zookeeper_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="tf")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill tf zookeeper were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK with TF\n")
            pytest.skip("HA tests to kill tf zookeeper were skipped ")

    def check_ha_kill_tf_zookeeper_nal(self):
        """Kill tf zookeeper-nal

        Executed only on MOSK cluster with TF

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found nodes where tf zookeeper-nal is running
            2. SSH to node
            3. Kill tf zookeeper-nal service
            4. Check tf zookeeper-nal agent become ready
            5. Check that all pods are Running and Ready

        Expected result - tf zookeeper-nal services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version and self.cluster.tf_enabled():
            nodes = self.cluster.get_machines(
                k8s_labels={"tfanalyticsdb": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_zookeeper_tf-zookeeper-nal-.?_tf'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                zookeeper_nal_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | grep /data/conf/zoo.cfg | cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {zookeeper_nal_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="tf")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill tf zookeeper-nal were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK with TF\n")
            pytest.skip("HA tests to kill tf zookeeper-nal were skipped ")

    def check_ha_kill_tf_rabbitmq(self):
        """Kill tf rabbitmq

        Executed only on MOSK cluster with TF

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found nodes where tf rabbitmq is running
            2. SSH to node
            3. Kill tf rabbitmq service
            4. Check tf rabbitmq agent become ready
            5. Check that all pods are Running and Ready

        Expected result - tf rabbitmq services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version and self.cluster.tf_enabled():
            nodes = self.cluster.get_machines(
                k8s_labels={"tfconfigdb": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_rabbitmq_tf-rabbitmq-.?_tf'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                rabbitmq_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | grep beam.smp | cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {rabbitmq_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="tf")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill tf rabbitmq were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK with TF\n")
            pytest.skip("HA tests to kill tf rabbitmq were skipped")

    def check_ha_kill_tf_redis(self):
        """Kill tf redis

        Executed only on MOSK cluster with TF

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found nodes where tf redis is running
            2. SSH to node
            3. Kill tf redis service
            4. Check tf redis agent become ready
            5. Check that all pods are Running and Ready

        Expected result - tf redis services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version and self.cluster.tf_enabled():
            nodes = self.cluster.get_machines(
                k8s_labels={"tfanalyticsdb": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_redis_redis-tf-redis-.?_tf'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                redis_pid = node._run_cmd(
                    f"docker container top {cont_name} -o pid,args | grep redis-server | cut -d ' ' -f 1",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {redis_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="tf")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill tf redis were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK with TF\n")
            pytest.skip("HA tests to kill tf redis were skipped")

    def check_ha_kill_libvirt(self):
        """Kill libvirt

        Executed only on MOSK cluster

        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. Found nodes where libvirt is running
            2. SSH to node
            3. Kill libvirt service
            4. Check libvirt agent become ready
            5. Check that all pods are Running and Ready

        Expected result - libvirt services restored successfully.
        """
        cluster_version = self.cluster.clusterrelease_version
        if settings.RUN_HA_TESTS and 'mosk' in cluster_version:
            nodes = self.cluster.get_machines(
                k8s_labels={"openstack-compute-node": "enabled"})
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                cont_name = node._run_cmd(
                    "docker ps --format '{{.Names}}' | grep -E '^k8s_libvirt_libvirt-libvirt'",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                libvirt_pid = node._run_cmd(
                    f"docker exec {cont_name} cat /var/run/libvirtd.pid",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                node._run_cmd(f"sudo kill -9 {libvirt_pid}", ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                LOG.info("Waiting for pods to be in a correct state")
                self.check_k8s_pods(timeout=1200, interval=30, target_namespaces="openstack")
        else:
            LOG.info("\nHA tests to kill libvirt were skipped "
                     "as RUN_HA_TESTS was set to False or child cluster isn't MOSK\n")
            pytest.skip("HA tests to kill libvirt were skipped")

    def check_ha_kill_system_proc_in_cont(self):
        """Kill main process in mke system containers
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to each cluster node 1 by 1
            2. Collect list of container id's
            3. Kill containers 1 by 1
            4. Wait till all containers are running again
            5. Check that all pods are in a correct state

        Expected result - pods and their containers are Running and Ready.
        """
        # see PRODX-15408 for details
        cont_to_skip = ['ucp-kubectl',
                        'ucp-cluster-agent']
        if settings.RUN_HA_TESTS:
            nodes = self.cluster.get_machines()
            KEY_FILE = settings.HA_TEST_PRIVATE_KEY_FILE
            for node in nodes:
                LOG.info(f"Access {node.name}")
                cont_list = node._run_cmd("docker ps --format '{{.Names}}' |"
                                          " grep -v k8s",
                                          ssh_key=KEY_FILE).stdout_str.split("\n")
                LOG.debug(f"Containers: {cont_list}")
                LOG.debug("The state of the containers before the start of the test")
                containers_before_ha = node._run_cmd("docker ps | grep -v k8s",
                                                     ssh_key=KEY_FILE).stdout_str
                LOG.debug(f"{containers_before_ha}")
                LOG.info("Sending kill -9 for each container")
                for container in cont_list:
                    _container = container.split(".")[0]
                    if _container not in cont_to_skip:
                        node._run_cmd(f"id=`docker ps | grep {_container} |"
                                      f"awk {{'print $1'}}`;"
                                      f"docker kill --signal=9 $id",
                                      ssh_key=KEY_FILE)
                        cmd = f"docker ps | grep {_container} |  wc -l"
                        waiters.wait(lambda:
                                     node._run_cmd(cmd, ssh_key=KEY_FILE).
                                     stdout_str == '1', timeout=60, interval=5)
                    else:
                        LOG.info(f"Container {_container} skipped because "
                                 f"it is in a skip list")
                LOG.debug("The state of the containers after the start of the test")
                containers_after_ha = node._run_cmd("docker ps | grep -v k8s",
                                                    ssh_key=KEY_FILE).\
                    stdout_str
                LOG.debug(f"{containers_after_ha}")
                LOG.info("Waiting for pods")
                self.check_k8s_pods(timeout=1200, interval=30)
        else:
            LOG.info("\nHA tests Kill processes in containers were skipped"
                     "as RUN_HA_TESTS was set to False\n")

    def check_ha_kill_ucp_cluster_agent(self):
        """Kill ucp-cluster-agent container
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to each cluster node 1 by 1
            2. Search for ucp-cluster-agent container
            3. Kill ucp-cluster-agent container
            4. Wait till container IS running
            5. Check that all pods are in a correct state
        Expected result - pods and container are Running and Ready.
        """

        def check_if_cluster_agent_is_running():
            for node in nodes:
                LOG.info(f"Access {node.name}")
                cont_count = \
                    node._run_cmd(f"docker ps | grep {container} | "
                                  f"wc -l", ssh_key=KEY_FILE).stdout_str
                if int(cont_count) == 1:
                    return True
                elif int(cont_count) > 1:
                    LOG.warning(f"Found {cont_count} ucp-cluster-agent containers "
                                f"on {node.name} - only 1 was expected.")
        if settings.RUN_HA_TESTS:
            nodes = self.cluster.get_machines()
            container = 'ucp-cluster-agent'
            KEY_FILE = settings.HA_TEST_PRIVATE_KEY_FILE
            for node in nodes:
                LOG.info(f"Access {node.name}")
                cont = node._run_cmd(f"docker ps | grep {container} | wc -l",
                                     ssh_key=KEY_FILE).stdout_str
                if int(cont) != 0:
                    LOG.info(f"Container {container} found. Killing it now")
                    node._run_cmd(f"id=`docker ps | grep {container} |"
                                  f" awk {{'print $1'}}`;"
                                  f"docker kill --signal=9 $id",
                                  ssh_key=KEY_FILE)
                    break
            else:
                raise Exception('ucp-cluster-agent was not found on any nodes')
            waiters.wait(lambda: check_if_cluster_agent_is_running(),
                         timeout=60, interval=10)
        self.check_k8s_pods(timeout=1200, interval=30)

    def check_ha_stop_containerd(self):
        """Stop containerd service on every cluster node (one at a time)
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to node
            2. Get number of containerd-shim processes
            3. Stop containerd service and make sure it is stopped
            4. Get number of containerd-shim processes and
               compare with previous value (must be the same)
            5. Check that all pods are Running and Ready
            6. Start containerd service

        Expected result - containerd-shim processes are running independently
        from containerd service and all containers in the pods are
        still operational even if containerd is down.
        """
        if settings.RUN_HA_TESTS:
            nodes = self.cluster.get_machines()
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                proc = node._run_cmd(
                    "pgrep -x containerd-shim | wc -l",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                LOG.info(f"There are containerd-shim {proc} processes")
                LOG.info("Stopping containerd")
                node._run_cmd("sudo service containerd stop",
                              ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                time.sleep(5)
                node._run_cmd("pgrep -x containerd",
                              check_exit_code=True,
                              expected_codes=[1],
                              ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                proc2 = node._run_cmd(
                    "pgrep -x containerd-shim | wc -l",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                LOG.info(f"Now it is {proc} containerd-shim processes")
                try:
                    assert int(proc) == int(proc2) + 1
                    self.check_k8s_pods(timeout=1200, interval=30)
                except Exception as e:
                    LOG.error(e)
                finally:
                    LOG.info("Start containerd")
                    node._run_cmd("sudo service containerd start",
                                  ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                # TODO: maybe check pod restarts also?
        else:
            LOG.info("\nHA tests to stop containerd-shim were skipped"
                     "as RUN_HA_TESTS was set to False\n")

    def check_ha_restart_docker_service(self):
        """Restart docker service on every cluster node (one at a time)
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to node
            2. Restart docker service
            3. Check docker services are OK
            4. Check that all pods are Running and Ready

        Expected result - all docker services are OK after
        docker was restarted.
        """
        if settings.RUN_HA_TESTS:
            nodes = self.cluster.get_machines()
            ucp_worker_agent_name = self.get_ucp_worker_agent_name()
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                node._run_cmd("sudo systemctl restart docker.service",
                              ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                time.sleep(60)
                LOG.info("Waiting for k8s pods to be ready")
                waiters.wait_pass(
                    lambda: self.k8sclient.pods.list_all(), timeout=120)
                LOG.info("Waiting for correct docker service replicas")
                # Check/wait for correct docker service replicas in cluster
                self.check_actual_expected_docker_services(
                    changed_after_upd={'ucp-worker-agent-x': ucp_worker_agent_name})
                self.check_k8s_pods(timeout=1200, interval=30)
        else:
            LOG.info("\nHA test to restart docker service were skipped "
                     "as RUN_HA_TESTS was set to False\n")

    def check_ha_haproxy_lb(self, ssh_user='mcc-user'):
        """Check that each node can balance traffic through HAProxy

        We have to check that all HAproxy can handle outage of any nodes/services

        Scenario:
            1. Select node to serve public ip. Do for each node
            2. Stop mcc-keepalived service on other nodes
            3. Check that API works well
            4. Pause k8s API docker service on other nodes
            5. Check that API is available
            6. Stop k8s API and mke API docker services to drop any connection to HAProxy
            7. Check that docker services were restored and k8s API is available
            8. Check that a pod can be scheduled on the current node
            9. Restore k8s API and keepalived on other nodes
            10. Check that API works well
        """
        if not settings.RUN_HA_TESTS or \
                self.cluster.provider not in (
                    utils.Provider.baremetal,
                    utils.Provider.vsphere,
                    utils.Provider.equinixmetal,
                    utils.Provider.equinixmetalv2):
            msg = ("\nHA test to check HAProxy LB service were skipped "
                   "as RUN_HA_TESTS was set to False or cluster doesn't support the feature")
            LOG.info(msg)
            pytest.skip(msg)
            return

        cluster_release = self.cluster.clusterrelease_version
        LOG.info(f"Cluster release: {cluster_release}")
        can_check_mke = 'mke' in cluster_release and \
                        version.parse(cluster_release) \
                        > version.parse('mke-7-5-0-3-4-6')
        can_check_mos = 'mos' in cluster_release and \
                        version.parse(cluster_release) \
                        > version.parse('mos-6-20-0-21-6')
        if not (can_check_mke or can_check_mos):
            LOG.info("\nHA test to check HAProxy LB service were skipped "
                     f"as cluster {cluster_release} version isn't support the feature\n")
            return

        ssh_key_file = settings.HA_TEST_PRIVATE_KEY_FILE
        nodes = self.cluster.get_machines(machine_type="control")
        try:
            LOG.info("Cluster has %s control nodes", [m.name for m in nodes])
            # 1. Select one control node
            for one in (nodes + [nodes[0]]):
                info = f"#  Check Haproxy on the node {one.name}  #"
                LOG.info(f"\n{'#' * len(info)}"
                         f"\n{info}"
                         f"\n{'#' * len(info)}")
                # Select other contronl nodes
                other_nodes = set(nodes) - set([one])
                # 2. Stop Keepalived on other nodes to move VIP to the selected control node
                LOG.info("Turning off Keepalived on %s", [m.name for m in other_nodes])
                for stop_one in other_nodes:
                    LOG.info("Stop keepalived service on %s", stop_one.name)
                    stop_one._run_cmd(
                        "sudo systemctl stop mcc-keepalived.service",
                        verbose=True,
                        ssh_key=ssh_key_file,
                        ssh_login=ssh_user,
                        reconnect=True)

                # 3. Check API
                self.cluster.check.check_k8s_nodes()
                self.cluster.check.check_actual_expected_pods()

                # 4. Turn off k8s API on other nodes
                LOG.info("Turning off backend k8s API services on %s", [m.name for m in other_nodes])
                for stop_one in other_nodes:
                    LOG.info("Pause k8s api service on %s", stop_one.name)
                    stop_one._run_cmd(
                        "sudo docker pause ucp-kube-apiserver",
                        verbose=True,
                        ssh_key=ssh_key_file,
                        ssh_login=ssh_user,
                        reconnect=True)

                LOG.info("Wait 10 seconds before HAProxy reacts to stop api services")
                time.sleep(10)

                # 5. Check API availability
                LOG.info("Getting list of nodes to check API availability")
                assert_nodes = self.cluster.get_machines(machine_type="control")

                assert set(n.name for n in nodes) == set(n.name for n in assert_nodes), \
                    ("Can't fetch list of nodes. It seems that haproxy can't handle "
                     f"service off on {other_nodes}")
                LOG.info(f"Response with machines: {[n.name for n in assert_nodes]}")

                # 6. Simulate haproxy <-> ucp-controller connection failure
                #    MKE containers must be restarted automatically in a few seconds
                one._run_cmd(
                    "sudo docker stop ucp-kube-apiserver; sudo docker stop ucp-controller",
                    verbose=True,
                    ssh_key=ssh_key_file,
                    ssh_login=ssh_user,
                    reconnect=True)

                # 7. Check that MKE containers were restarted and k8s API is available
                LOG.info(f"\n\nWait until k8s API becomes available again "
                         f"and VIP is returned on the machine {one.name}\n")
                waiters.wait_pass(
                    self.cluster.get_machines, timeout=600, interval=10,
                    expected=(ApiException, MaxRetryError, ProtocolError),
                    timeout_msg=f"MKE services were not automatically restarted on the machine {one.name}")

                LOG.info("\n\n*** VIP is restored ***\n")
                LOG.info("Get the list of machines to check API availability after stopping MKE services")
                assert_nodes = self.cluster.get_machines(machine_type="control")
                assert set(n.name for n in nodes) == set(n.name for n in assert_nodes), \
                    ("Can't fetch list of machines. It seems that haproxy can't handle "
                     f"service off on {other_nodes}")
                LOG.info(f"Response with machines: {[n.name for n in assert_nodes]}")

                # 8. Check that a pod can be scheduled on the current node
                #    Use exec_pod_cmd() to run a simple pod
                one.exec_pod_cmd("docker ps|grep ucp-kube-apiserver; sudo ctr tasks ls;", verbose=True)

                # 9. Restore k8s API and keepalived on all nodes
                LOG.info("Recover k8s api services on %s", [m.name for m in other_nodes])
                for start_one in other_nodes:
                    LOG.info("unpause k8s api service on %s", start_one.name)
                    start_one._run_cmd(
                        "sudo docker unpause ucp-kube-apiserver",
                        verbose=True,
                        ssh_key=ssh_key_file,
                        ssh_login=ssh_user,
                        reconnect=True)

                self.cluster.check.check_k8s_nodes()
                self.cluster.check.check_actual_expected_pods()

                LOG.info("Turning on Keepalived on %s", [m.name for m in other_nodes])
                for start_one in other_nodes:
                    LOG.info("Start Keepalived service on %s", start_one.name)
                    start_one._run_cmd(
                        "sudo systemctl start mcc-keepalived.service",
                        verbose=True,
                        ssh_key=ssh_key_file,
                        ssh_login=ssh_user,
                        reconnect=True)

                LOG.info("Wait 30 seconds before HAProxy reacts to start api services. "
                         "Should give time to restore clustered pods.")
                time.sleep(30)

                # 10. Check API
                self.cluster.check.check_k8s_nodes()
                self.cluster.check.check_actual_expected_pods()

                LOG.info("Finish checking HAProxy service on %s", one.name)
        finally:
            LOG.info("\n#########################################################"
                     "\n#  Restore k8s and keepalived services on all machines  #"
                     "\n#########################################################")
            for one in nodes:
                LOG.info(f"Unpause k8s service on the node {one.name}")
                one._run_cmd(
                    "sudo docker unpause ucp-kube-apiserver || true",
                    verbose=True,
                    ssh_key=ssh_key_file,
                    ssh_login=ssh_user,
                    reconnect=True)
                LOG.info(f"Start Keepalived service on the node {one.name}")
                one._run_cmd(
                    "sudo systemctl start mcc-keepalived.service || true",
                    verbose=True,
                    ssh_key=ssh_key_file,
                    ssh_login=ssh_user,
                    reconnect=True)
            self.cluster.check.check_k8s_nodes()
            self.cluster.check.check_actual_expected_pods()

    def check_ha_ceph_mgr(self):
        """Check that Ceph HA move active mgr to "stand by" node

        Scenario:
            1. Check ceph cluster state and readiness
            2. Check on which node we have active ceph mgr and on which it in stand by
            3. Reboot node with active ceph mgr
            4. Check that all pods are Running and Ready
            5. Check ceph cluster state and readiness
            6. Check that ceph mgrs swapped

        Expected result - Ceph active mgr should migrate to node on which it was
        in stand by state
        """
        if not settings.RUN_HA_TESTS or \
                self.cluster.provider not in (
                    utils.Provider.baremetal,
                    utils.Provider.equinixmetalv2):
            msg = ("\nHA test to check HA  Ceph mgr service were skipped "
                   "as RUN_HA_TESTS was set to False or cluster doesn't support the feature")
            LOG.info(msg)
            pytest.skip(msg)

        cluster_version = self.cluster.clusterrelease_version
        LOG.info(f'{cluster_version}, and parse : {version.parse(cluster_version)}')
        if 'mke-11' in cluster_version and version.parse(cluster_version) < version.parse("mke-11-6-0-3-5-5"):
            pytest.skip(f'Skip test for cluster {cluster_version}')

        if 'mke-13' in cluster_version and version.parse(cluster_version) < version.parse('mke-13-2-5-3-5-2'):
            pytest.skip(f'Skip test for cluster {cluster_version}')

        if 'mosk' in cluster_version and version.parse(cluster_version) < version.parse('mosk-12-6-0-rc-23-1'):
            pytest.skip(f'Skip test for cluster {cluster_version}')

        def check_active_mgr_similar(cluster, ceph_tools_pod):
            mgr_status = cluster.get_cephcluster().data.get('status', {}) \
                .get('fullClusterInfo', {}).get('daemonsStatus', {}).get('mgr', {}).get('status', '')
            if not mgr_status:
                LOG.warning('Ceph mgr status field is empty')
                return False
            active_mgr_idx = mgr_status.split(' ')[0]

            # Also check status in mgr pod and compare it
            cmd = ['/bin/sh', '-c', 'ceph mgr dump --format json']
            json_ret = ceph_tools_pod.exec(cmd)
            mgr_in_pod = yaml.safe_load(json_ret).get('active_name', '')
            if not mgr_in_pod:
                LOG.warning('mgr name in pos is empty')
                return False

            if mgr_in_pod == active_mgr_idx:
                LOG.info(f'Ceph mgr name in pod = "{mgr_in_pod}" is same as in kcc: "{active_mgr_idx}"')
                return True
            else:
                return False

        ceph_health_timeout = 2400
        rook_ns = settings.ROOK_CEPH_NS
        ssh_key_file = utils.load_keyfile(settings.HA_TEST_PRIVATE_KEY_FILE)
        pkey = utils.get_rsa_key(ssh_key_file['private'])
        ceph_tools_pod = self.cluster.get_ceph_tool_pod()
        #
        # 1. Check ceph cluster state and readiness
        LOG.info("Check ceph cluster status")
        self.cluster.check.wait_ceph_health_status(
            timeout=ceph_health_timeout, interval=30)
        self.cluster.check.check_cluster_readiness()
        self.cluster.check.check_ceph_pvc()
        LOG.info("Ceph cluster is healthy.")

        # 2. Check on which node we have active ceph mgr and on which it in stand by
        ceph_mgr_pods = self.k8sclient.pods.list(
            namespace=rook_ns,
            name_prefix='rook-ceph-mgr')
        # Compare mgr status in pod and in kcc
        waiters.wait(lambda: check_active_mgr_similar(self.cluster, ceph_tools_pod),
                     timeout=600,
                     interval=10,
                     timeout_msg='Expected that values will be similar in pod and in kcc')
        # Calculate active mgr node
        ceph_mgr_status = self.cluster.get_cephcluster().data.get('status', {})\
            .get('fullClusterInfo', {}).get('daemonsStatus', {}).get('mgr', {}).get('status')
        initial_active_mgr_idx = ceph_mgr_status.split(' ')[0]
        pod_with_active_mgr = [pod for pod in ceph_mgr_pods if f'rook-ceph-mgr-{initial_active_mgr_idx}' in pod.name][0]
        LOG.info(f'Current pod name with active Ceph mgr: {pod_with_active_mgr.name}')
        # Find node on which active mgr works
        node_with_active_mgr_name = pod_with_active_mgr.data.get('spec', {}).get('node_name', '')
        assert node_with_active_mgr_name, 'node_name field is empty'

        node_with_active_mgr_address = pod_with_active_mgr.data.get('status', {}).get('host_ip', '')
        assert node_with_active_mgr_address, 'host_ip field is empty'

        LOG.info(f'Node with active Ceph mgr pod: {node_with_active_mgr_name} '
                 f'with IP address: {node_with_active_mgr_address}')
        machine = [machine for machine in self.cluster.get_machines() if machine.get_k8s_node_name() ==
                   node_with_active_mgr_name][0]

        # 3. Reboot node with active ceph mgr
        LOG.info(f'Accessing {machine.name} node and rebooting it')
        auth = exec_helpers.SSHAuth(username='mcc-user',
                                    password='', key=pkey)
        ssh = exec_helpers.SSHClient(
            host=node_with_active_mgr_address, port=22, auth=auth)
        ssh.logger.addHandler(logger.console)
        ssh.sudo_mode = True
        ssh.reconnect()
        ssh.check_call("iptables -A INPUT -p tcp --dport 22 -j REJECT; "
                       "/sbin/shutdown -r -f now &>/dev/null & exit")
        LOG.info("Node is being rebooted.")
        waiters.wait_tcp(node_with_active_mgr_address, port=22, timeout=600)
        LOG.info("Node is rebooted succesfully. Establishing new connection...")
        ssh = exec_helpers.SSHClient(host=node_with_active_mgr_address, port=22, auth=auth)
        ssh.logger.addHandler(logger.console)

        # 4. Check that all pods are Running and Ready
        LOG.info("Check cluster nodes and pods")
        self.cluster.check.check_k8s_nodes()
        self.cluster.check.check_actual_expected_pods(timeout=1800)

        # 5. Check ceph cluster state and readiness
        LOG.info("Check ceph cluster status")
        self.cluster.check.wait_ceph_health_status(
            timeout=ceph_health_timeout, interval=30)
        self.cluster.check.check_cluster_readiness()
        self.cluster.check.check_ceph_pvc()
        LOG.info("Ceph cluster is healthy.")

        # 6. Check that ceph mgrs swapped
        # Compare mgr status in pod and in kcc
        waiters.wait(lambda: check_active_mgr_similar(self.cluster, ceph_tools_pod),
                     timeout=600,
                     interval=10,
                     timeout_msg='Expected that values will be similar in pod and in kcc')

        ceph_mgr_status = self.cluster.get_cephcluster().data.get('status', {}) \
            .get('fullClusterInfo', {}).get('daemonsStatus', {}).get('mgr', {}).get('status')
        last_active_mgr_idx = ceph_mgr_status.split(' ')[0]
        assert initial_active_mgr_idx != last_active_mgr_idx, f'Initial active manager pod index: ' \
                                                              f'"{initial_active_mgr_idx}" is same with last ' \
                                                              f'active manager pod index: "{last_active_mgr_idx}'""
        LOG.info(f'HA test for Ceph mgr pod finished successfully!, initial pod was: "{initial_active_mgr_idx}"'
                 f'and new pod is: "{last_active_mgr_idx}"')

    def calculate_k8s_downtime(self, data: dict) -> list:
        """
        Calculate potential downtime period based on Prometheus data

        Args:
            data: PrometheusClient.get_svc_probe_duration_seconds object

        Returns: result list of downtimes

        """
        result = []
        start = {}
        stop = {}
        for item in data[0].get('values', []):
            if item[1] == "0":
                if not start:
                    start = item[0]
                else:
                    stop = item[0]
            else:
                if start and stop:
                    time_format = '%H:%M:%S'
                    start_data = datetime.fromtimestamp(start, tz=timezone.utc)
                    stop_data = datetime.fromtimestamp(stop, tz=timezone.utc)
                    duration = stop_data - start_data

                    result.append({
                        "start": start_data.strftime(time_format),
                        "stop": stop_data.strftime(time_format),
                        "duration": duration
                    })
                start = {}
                stop = {}
        return result

    def check_svc_downtime(self, service_name, expected_downtime, namespace=""):
        """
        Check downtime for k8s services
        Args:
            service_name: service name
            expected_downtime: expected downtime in seconds
            namespace: service namespace

        Returns: None

        """
        start = time.time()
        # TODO delete kaas release version check after 2.23 release
        if self.cluster.sl_ha_enabled() and \
                version.parse(self.cluster.get_kaasrelease_version()) >= version.parse("kaas-2-23-0-rc"):
            expected_downtime = settings.HA_SL_SVC_HA_MODE_DOWNTIME
            LOG.info("SL HA enabled")
        else:
            LOG.info("SL HA disabled")

        svc_pods = self.k8sclient.pods.list_starts_with(service_name, namespace=namespace)
        target_pod = svc_pods[0]
        LOG.info(f"Delete {target_pod.name} pod")
        target_pod.delete()
        LOG.info(f"Wait until all {service_name} pod(s) Created and Running")
        self.check_k8s_pods(pods_prefix=service_name,
                            target_namespaces=namespace,
                            timeout=300,
                            interval=30)
        end = time.time()
        client = self.cluster.prometheusclient
        result = client.get_svc_probe_success(namespace=namespace, service_name=service_name, start=start, end=end)
        downtime = self.calculate_k8s_downtime(result)
        self.check_downtime(downtime, expected_downtime)

    def check_downtime(self, downtime, expected_downtime: int = settings.REFAPP_EXPECTED_DOWNTIME):
        """
        Check is existing downtime in list more than expected
        Args:
            expected_downtime: time in seconds
            downtime: list of downtime objects

        Returns: None

        """

        if downtime:
            LOG.error("The following downtime ranges(s) found:\n" + tabulate(downtime, tablefmt="presto",
                                                                             headers="keys"))
        not_expected_downtime = list(filter(lambda x: x['duration'] > timedelta(seconds=expected_downtime), downtime))
        if not_expected_downtime:
            LOG.error(f"The following downtime ranges(s) more that expected ({expected_downtime} s.)\n"
                      + tabulate(not_expected_downtime, tablefmt="presto", headers="keys"))
            raise Exception("Current downtime during testing is longer than expected")
        else:
            LOG.info("Current downtime within expected limits")

    def check_helmbundles(self, expected_spec=None, timeout=1800, interval=60):
        def helmbundles_ready(expected_spec):
            try:
                hb = self.cluster.get_helmbundle().data
                expected_spec = expected_spec or hb['spec']['releases']
            except Exception as e:
                LOG.error(e)
                return False

            hb_status = hb.get('status') or {}
            release_status = hb_status.get('releaseStatuses', {})

            spec_chart_names = ["{0}:{1}".format(
                                x['name'],
                                helpers.get_chart_version(x))
                                for x in expected_spec]
            status_chart_names = ["{0}:{1}".format(k, v['version'])
                                  for k, v in release_status.items()]
            LOG.debug(f"Current status "
                      f"of the charts:\n{yaml.dump(release_status)}")

            # Check that all charts with expected versions
            # are present in the helmbundle status
            charts_diff = set(spec_chart_names)\
                .difference(set(status_chart_names))
            if charts_diff:
                LOG.error(f"Mismatch between expected and actual helmbundles. "
                          f"Expected: {spec_chart_names}, "
                          f"actual: {status_chart_names}, "
                          f"difference: {charts_diff}")
                return False

            # Check that no charts in the helmbundle status has False
            failed_charts = [v for v in release_status.values()
                             if not v['success']]
            if failed_charts:
                LOG.error(f"The following charts "
                          f"in the failed state:\n"
                          f"{yaml.dump(failed_charts)}")
                return False

            LOG.info(f"The following charts "
                     f"have been deployed "
                     f"in the cluster:\n{yaml.dump(status_chart_names)}")
            return True

        LOG.info("Check helmbundles status")
        try:
            waiters.wait(lambda: helmbundles_ready(expected_spec),
                         timeout=timeout, interval=interval)
        except exceptions.TimeoutError:
            helmbundle_status = self.cluster.get_helmbundle().data.get('status') or {}
            releaseStatuses = helmbundle_status.get('releaseStatuses', {})
            err = f"Not all helmbundles have been deployed: {releaseStatuses}"
            raise TimeoutError(err)
        LOG.info("All charts have been successfully deployed")

    def check_svc_access(self, remote):
        tmpl = templates.render_template(settings.SVC_CHECK_POD_YAML)
        LOG.debug(tmpl)
        json_body = json.dumps(yaml.load(tmpl, Loader=yaml.SafeLoader))
        LOG.info("Creating nginx pod")
        pod = self.cluster.k8sclient.pods.create(
            name="svc-check", namespace="default",
            body=json.loads(json_body)
        )
        pod.wait_phase('Running', timeout=1800, interval=10)
        LOG.info("Creating service")
        service = kubernetes.client.V1Service()
        service.metadata = kubernetes.client.V1ObjectMeta(
            name="svc-check-svc")
        service.spec = kubernetes.client.V1ServiceSpec(type="LoadBalancer")
        service.spec.selector = {"run": "svc-check"}
        service.spec.ports = [
            kubernetes.client.V1ServicePort(port=80, target_port=80)
        ]
        svc = self.cluster.k8sclient.services.create(
            name="svc-check-svc", namespace='default', body=service)
        LOG.info("Waiting for LB to have external ip/hostname")
        waiters.wait(lambda: svc.read().to_dict()['status'][
            'load_balancer']['ingress'] is not None,
            timeout=500, interval=15,
            timeout_msg='Timeout waiting for svc loadbalancer')
        ingress_addr = \
            svc.read().to_dict()['status']['load_balancer']['ingress'][0]
        if ingress_addr['ip']:
            address = ingress_addr['ip']
        else:
            address = ingress_addr['hostname'] or ''
        LOG.info(f"LB iaddress is {address}")

        command = f'wget -O- --no-proxy {address}' \
            if settings.KAAS_OFFLINE_DEPLOYMENT else f'wget -O- {address}'
        ret = remote.check_call(command, raise_on_err=False)
        pod.delete()
        svc.delete()
        assert ret.exit_code == 0 and "Welcome to nginx" in ret.stdout_str,\
            f"Service is not available. Response {ret.stdout_str}"

    def check_resource_discovery(self):
        if not self.cluster.is_management and not self.cluster.is_regional:
            LOG.warning("check_resource_discovery() is not suitable "
                        "for child clusters")
            return None
        if self.cluster.provider == utils.Provider.aws:
            LOG.info("Checking aws resource discovery")
            kaas_awsresources = \
                self.cluster.k8sclient.kaas_awsresources.list(
                    namespace=self.cluster.namespace)
            if self.cluster.is_management:
                awsresources = [x for x in kaas_awsresources
                                if x.name == 'cloud-config'][0]
            else:
                awsresources = [x for x in kaas_awsresources
                                if x.name == self.cluster.name][0]
            aws_region = self.cluster.data['spec']['providerSpec'][
                'value']['region']
            regions = awsresources.regions
            LOG.debug(f"AWS regions: {regions}")
            instance_types = awsresources.instance_types(aws_region)
            LOG.debug(f"AWS instance types: {instance_types}")
            images = awsresources.images(aws_region)
            LOG.debug(f"AWS images: {images}")
            assert regions, \
                f"No regions were found in " \
                f"{awsresources.name} kaas_awsresources crd object"
            assert instance_types, \
                f"No instance_types were found in " \
                f"{awsresources.name} kaas_awsresources crd object"
            assert images, \
                f"No images were found in " \
                f"{awsresources.name} kaas_awsresources crd object"
        else:
            LOG.info("Skipping check_resource_discovery")

    def check_machines_status(self, expected_status='Ready',
                              timeout=settings.SI_CHECK_MACHINES_STATUS_TIMEOUT, interval=30):
        def machines_status(expected_status):
            machines = self.cluster.get_machines()
            LOG.debug(f"Got machines: {machines}")
            machines_in_status = [x for x in machines
                                  if x.machine_status == expected_status]
            LOG.debug(f"Machines in {expected_status} "
                      f"status: {machines_in_status}")
            if len(machines) != len(machines_in_status):
                for machine in [x for x in machines
                                if x not in machines_in_status]:
                    LOG.info(f"{machine.name} machine is "
                             f"in status {machine.machine_status}")
                LOG.info(f"===Not all machines are "
                         f"in status {expected_status}===")
                raise Exception(f"Not all machines are "
                                f"in status {expected_status}")
            return True

        LOG.info(f"Wait {timeout} seconds until machines will be in "
                 f"the status <{expected_status}>")
        waiters.wait_pass(lambda: machines_status(expected_status),
                          timeout=timeout, interval=interval,
                          expected=(Exception, ApiException))
        LOG.info("All machines are in expected status")

    def check_any_machine_exp_status(self, expected_status='Ready',
                                     timeout=1800, interval=30):
        def machines_status(exp_status):
            machines = self.cluster.get_machines()
            LOG.debug(f"Got machines: {machines}")
            machines_in_status = [x for x in machines
                                  if x.machine_status == exp_status]
            LOG.info(f"Machines in {exp_status} "
                     f"status: {machines_in_status}")
            if not machines_in_status:
                raise Exception(f"There are not any machines "
                                f"in status {exp_status}")
            return True

        LOG.info(f"Wait {timeout} seconds until any of machines will be in "
                 f"the status <{expected_status}>")
        waiters.wait_pass(lambda: machines_status(expected_status),
                          timeout=timeout, interval=interval,
                          expected=(Exception, ApiException))
        LOG.info("There is machine in expected status")

    def check_kubeconfig_stored_in_userdata(self, one_machine_check=True,
                                            machines_list=None):
        # Since 2.12 kubeconfig stored in user-data should not be valid

        all_machines = machines_list or self.cluster.get_machines()
        machines = []
        if one_machine_check:
            machines.append(random.choice(all_machines))
        else:
            machines = all_machines
        for machine in machines:
            result = machine.exec_pod_cmd(
                "sudo cat /var/lib/cloud/instance/user-data.txt",
                verbose=False)['logs']
            userdata = yaml.safe_load(result)
            for u in userdata['write_files']:
                if u['path'] == '/etc/lcm/kubeconfig':
                    kubeconfig = u['content']
                    with open(f'kubeconfig-{machine.name}', 'w') as f:
                        f.write(kubeconfig)
            try:
                # Init a new Manager instance with the specified kubeconfig file
                kmanager = self.cluster._manager.__class__(kubeconfig=f'kubeconfig-{machine.name}')
                self_lcmmachine = kmanager.get_lcmmachine(
                    name=machine.name,
                    namespace=self.cluster.namespace)
                lcmm_name = self_lcmmachine.read().to_dict()['metadata']['name']
                if lcmm_name == machine.name:
                    raise Exception("kubeconfig taken from the user-data of "
                                    f"node {lcmm_name} should not have access "
                                    "to the mgmt cluster (its own lcmmachine)")
            except ApiException as e:
                if e.reason == 'Unauthorized':
                    LOG.info(f"Machine {machine.name} has not access to its "
                             "own lcmmachine in mgmt cluster as expected")
                    return True

    def check_machines_access_with_given_key(self,
                                             machines_list=None,
                                             verbose=False,
                                             ssh_key_file=None,
                                             ssh_login='mcc-user'):
        machines = machines_list if machines_list \
            else self.cluster.get_machines()
        ssh_key_file = ssh_key_file \
            if ssh_key_file else self.cluster.private_key
        machines_names = [m.name for m in machines]
        not_ready_machines = []
        for machine in machines:
            try:
                machine._run_cmd("hostname", verbose=verbose,
                                 ssh_key=ssh_key_file,
                                 ssh_login=ssh_login,
                                 reconnect=True)
            except Exception as e:
                LOG.debug("Can't connect to {}: {}".format(
                    machine.name, e))
                not_ready_machines.append(machine.name)
        if not_ready_machines:
            LOG.warning("No access to next machines: {} "
                        "using key {}".format(
                            not_ready_machines, ssh_key_file))
            return False
        LOG.info("Succesfully connected to all machines: {} "
                 "using key {}".format(
                     machines_names, ssh_key_file))
        return True

    def check_update_finished(self, timeout=9300, interval=15,
                              check_managed_clusters=False,
                              expected_status="Ready"):
        def update_check(expected_status):
            machines_status_ok = True
            lcmmachines_status_ok = True
            clusterrelease_status_ok = True

            previous_clusterrelease = None
            status = self.cluster.data['status']
            if status:
                previous_clusterrelease = status.get('providerStatus', {}).get(
                    'releaseRefs', {}).get('previous', {}).get('name', {})
            spec_release = \
                self.cluster.spec['providerSpec']['value']['release']
            LOG.debug(f"Expected clusterrelease - {spec_release}, "
                      f"status.providerStatus.releaseRefs.previous.name - "
                      f"{previous_clusterrelease}")
            if spec_release != previous_clusterrelease:
                LOG.info(f"Expected clusterrelease - {spec_release}, "
                         f"actual clusterrelease - "
                         f"{previous_clusterrelease}. "
                         "Updating is not finished yet..")
                clusterrelease_status_ok = False

            if check_managed_clusters:
                # mgmt cluster has access to all machines,
                # so we check their status
                if self.cluster.is_management:
                    machines = \
                        self.k8sclient.kaas_machines.list_raw()
                    for machine in machines.to_dict()['items']:
                        m_name = machine['metadata']['name']
                        m_ns = machine['metadata']['namespace']
                        m_labels = machine['metadata'].get('labels', {})
                        cl_name = None if not m_labels else \
                            m_labels.get('cluster.sigs.k8s.io/cluster-name',
                                         None)
                        m_ps = machine['status'].get('providerStatus', {})
                        m_status = None if not m_ps else m_ps.get('status',
                                                                  None)
                        # we check that regional and mgmt machines are in
                        # expected_status otherwise we continue to wait
                        # CLUSTER_NAMESPACE is usually default and we expect
                        # regional cluster in this ns as well
                        if m_ns == settings.CLUSTER_NAMESPACE \
                                and m_status != expected_status:
                            LOG.info(f"Found {m_name} machine of {cl_name} "
                                     f"cluster in {m_status} status")
                            machines_status_ok = False
                        # we check that all machines, except mgmt
                        # and regional ones are in expected_status
                        if m_ns != settings.CLUSTER_NAMESPACE \
                                and m_status != expected_status:
                            LOG.error(f"Machine {m_name} of {cl_name} cluster "
                                      f"has status {m_status} while "
                                      f"Ready is expected")
                            raise Exception(f"Machine {m_name} has status "
                                            f"{m_status}")

                # mgmt & regional clusters have access to regional lcmmachines,
                # so we check their status here
                lcmmachines = \
                    self.cluster.get_all_lcmmachines(
                        states=["Upgrade", "Pending", "Deploy", "Prepare"]
                    )
                lcmm_not_ready = {}
                for lcmm in lcmmachines:
                    if lcmm.namespace != self.cluster.namespace:
                        lcmm_not_ready[lcmm.name] = {'ns': lcmm.namespace}
                    else:
                        LOG.info(f"{lcmm.name} lcmmachine of current "
                                 f"cluster is not Ready")
                        lcmmachines_status_ok = False
                if lcmm_not_ready:
                    LOG.error(f"These managed cluster lcmmachines "
                              f"are not ready: {lcmm_not_ready}")
                    raise Exception(f"These regional or child lcmachines "
                                    f"are not ready: {lcmm_not_ready}")
            else:
                # check lcmmachines that belong to current cluster
                my_lcmms = self.cluster.get_lcmmachines(
                    states=["Upgrade", "Pending", "Deploy", "Prepare"]
                )
                if my_lcmms:
                    lcmmachines_status_ok = False
                    LOG.info(f"Not Ready lcmmachines: {my_lcmms}")

                # check machines that belong to current cluster
                machines = self.cluster.get_machines()
                machines_in_status = [x for x in machines
                                      if x.machine_status == expected_status]
                LOG.debug(f"Machines in {expected_status} "
                          f"status: {machines_in_status}")
                if len(machines) != len(machines_in_status):
                    machines_status_ok = False
                for machine in [x for x in machines
                                if x not in machines_in_status]:
                    LOG.info(f"{machine.name} machine is "
                             f"in status {machine.machine_status}")

            if not machines_status_ok or not lcmmachines_status_ok:
                LOG.info(f"===Not all machines or lcmmachines are in "
                         f"status {expected_status}===")
                raise RuntimeError(f"Not all machines or lcmmachines "
                                   f"are in status {expected_status}")
            if not clusterrelease_status_ok:
                LOG.info("===Clusterrelease version of cluster hasn't "
                         "been updated===")
                raise RuntimeError("Clusterrelease version of cluster hasn't "
                                   "been updated")
            return True

        LOG.info(f"Waiting {timeout} sec until cluster is fully updated")
        waiters.wait_pass(lambda: update_check(expected_status),
                          timeout=timeout, interval=interval,
                          expected=(RuntimeError, ApiException,
                                    KeyError, TypeError, MaxRetryError))
        # we need to refresh cached expected pods/objects
        # lists in cluster object
        self.cluster._refresh_expected_objects()
        LOG.info("Update finished")

    def check_persistent_volumes_mounts(self, timeout=900, interval=10):
        def check_volumes_mounts():
            LOG.info("Checking local-volume-provisioner mounts!")
            hb = self.cluster.get_helmbundle().data
            releases = hb['spec']['releases']
            chart_name = [
                x['name'] for x in releases
                if x.get('chart') == 'kaas-bm/local-volume-provisioner'
            ]
            mounts = [x['values']['classes'] for x in releases
                      if x.get('chart') == 'kaas-bm/local-volume-provisioner']
            # Check if the chart present (for kaas < 1.13.0(2.0.0))
            if not chart_name:
                LOG.warning("LVP chart is not used for current release!")
                return True
            else:
                assert mounts, "Missing mounts!"
                LOG.info("Expected mounts on the nodes:")
                mount_paths = []
                for mount in mounts[0]:
                    mount_paths.append(mount['hostDir'])
                    LOG.info(mount['hostDir'])
                if not mount_paths:
                    LOG.warning("Local volumes mounts are not described"
                                "(or not used) in hb! Skip checking")
                else:
                    lvp_pods_list = \
                        self.cluster.k8sclient.pods.list_starts_with(
                            pattern=chart_name[0])
                    missing_mounts = []
                    for pod in lvp_pods_list:
                        LOG.info(f"Check mounts from pod: {pod.name}")
                        for path in mount_paths:
                            LOG.info(f"Mount: '{path}'")
                            if not pod.check_folder_exists(dir_path=path):
                                missing_mounts.append(path)
                        if missing_mounts:
                            raise FileExistsError(
                                f"Mount {missing_mounts} not found "
                                f"on the host!")
                return True

        waiters.wait_pass(lambda: check_volumes_mounts(),
                          timeout=timeout, interval=interval,
                          expected=(MaxRetryError, ApiException))

    def wait_machine_status_by_name(self, machine_name, expected_status,
                                    timeout=1800, interval=30):
        def machine_status(name):
            machine = self.cluster.get_machine(name)
            if not machine:
                LOG.warning(f"Machine {name} not found")
                return None
            status = machine.machine_status
            LOG.info(f"Machine {name} has <{status}> status")
            return status

        LOG.info(f"Wait {timeout} seconds until machine {machine_name} "
                 f"will be in the status <{expected_status}>")
        waiters.wait(lambda: machine_status(machine_name) == expected_status,
                     timeout=timeout, interval=interval)
        LOG.info(f"{machine_name} is in expected status {expected_status}")

    def check_boot_from_volume_by_machine_name(self, machine_name,
                                               openstack_client,
                                               boot_volume_size=80):
        # Check that bootFromVolume option still in machine CRD
        machine = self.cluster.get_machine(machine_name)
        current_boot_from_volume = machine.data.get('spec').get('providerSpec').get('value').get(
            'bootFromVolume', {}).get('enabled')
        assert current_boot_from_volume is True, f"Provided boot_from_volume: True flag " \
                                                 f"is not the same as in machine CRD: " \
                                                 f"{current_boot_from_volume}"

        current_boot_volume_size = machine.data.get('spec').get('providerSpec').get('value').get(
            'bootFromVolume', {}).get('volumeSize')
        assert boot_volume_size == current_boot_volume_size, f"Provided boot_volume_size: {boot_volume_size} " \
                                                             f"is not same as in machine CRD: " \
                                                             f"{current_boot_volume_size}"

        # Check os image from openstack client, if bootFromVolume worked correctly,
        # it should be empty
        openstack_machine_name = self.cluster.get_machine_openstack_name(machine_name)
        openstack_machine = openstack_client.get_server_by_name(openstack_machine_name)
        if openstack_machine.image:
            raise Exception(f"Image field for machine: {machine_name} is NOT empty, but should be, "
                            f"and contains next image(s): {openstack_machine.image}")

        LOG.info(f"All checks on machine: {machine_name} for bootFromVolume option done successfully, "
                 f"CRD is correct, openstack image is empty")

    def check_cluster_release(self, expected_clusterrelease,
                              timeout=1000, interval=30):

        if expected_clusterrelease:
            LOG.info(f"Checking cluster clusterrelease version. "
                     f"Expected: {expected_clusterrelease}")
            try:
                ec = expected_clusterrelease
                waiters.wait(
                    lambda: self.cluster.clusterrelease_version == ec,
                    timeout=timeout,
                    interval=interval
                )
            except exceptions.TimeoutError:
                raise TimeoutError(
                    f"Clusterrelease version mismatch. "
                    f"Actual: {self.cluster.clusterrelease_version}, "
                    f"expected: {expected_clusterrelease}")
            LOG.info("Clusterrelease version is correct")

    def check_actual_expected_pods(self,
                                   expected_pods=None,
                                   timeout=300,
                                   interval=10,
                                   exclude_pods=None,
                                   exclude_jobs=True,
                                   exclude_removed_nodes=True,
                                   exclude_incorrect_affinity=True,
                                   check_all_nss=False):
        """Compare expected list of pods (which is fetched automatically,
           unless explicitly provided) with actual list of pods in this
           cluster. Comparison is conducted for all namespaces by default"""

        LOG.info("Checking that all pods and their replicas are in place")

        def compare(expected_pods_dict,
                    exclude_jobs):
            failed = {}
            namespaces = expected_pods_dict.keys()
            actual_list = [pod for pod in
                           self.k8sclient.pods.list_raw().to_dict()['items']]

            if exclude_jobs:
                before_filtering = \
                    set([x['metadata']['name'] for x in actual_list])
                actual_list = \
                    [x for x in actual_list if not (
                        x['metadata']['owner_references'] and
                        x['metadata'][
                            'owner_references'][0]['kind'] == 'Job')]
                after_filtering = \
                    set([x['metadata']['name'] for x in actual_list])
                LOG.debug(f"These pods are jobs and will be filtered out: "
                          f"{before_filtering - after_filtering}")

            if exclude_removed_nodes:
                # After cluster scale down, some pods may be still mapped on the
                # removed nodes. Garbage collector may clean these pods in few hours
                # but we can ignore such pods here
                node_names = [node.name for node in self.k8sclient.nodes.list_all()]
                before_filtering = set([x['metadata']['name'] for x in actual_list])

                new_actual_list = []
                filtered_nodes = set()
                for pod in actual_list:
                    pod_name = f"{pod['metadata']['namespace']}/{pod['metadata']['name']}"

                    node_name = pod['spec']['node_name']
                    if node_name is not None and node_name not in node_names:
                        LOG.warning(f"Pod {pod_name} node name is bound to non-existing node {node_name}, "
                                    f"skip check the pod")
                        filtered_nodes.add(node_name)
                        continue

                    node_selector = pod['spec']['node_selector'] or {}
                    node_selector_hostname = node_selector.get('kubernetes.io/hostname')
                    if node_selector_hostname is not None and node_selector_hostname not in node_names:
                        LOG.warning(f"Pod {pod_name} node selector is bound to non-existing hostname "
                                    f"{node_selector_hostname}, skip check the pod")
                        filtered_nodes.add(node_selector_hostname)
                        continue

                    new_actual_list.append(pod)
                actual_list = new_actual_list
                filtered_nodes = list(filtered_nodes)

                after_filtering = set([x['metadata']['name'] for x in actual_list])

                if before_filtering - after_filtering:
                    LOG.warning(f"These pods will be ignored because assigned to non-existing nodes "
                                f"{filtered_nodes} : {before_filtering - after_filtering}")

            if exclude_incorrect_affinity:
                # After changing node labels, some pods may no longer match the node affinity
                before_filtering = set([x['metadata']['name'] for x in actual_list])

                actual_list = [x for x in actual_list
                               if not (x['status'].get('phase', '') == 'Failed'
                                       and x['status'].get('reason') == 'NodeAffinity')]

                after_filtering = set([x['metadata']['name'] for x in actual_list])

                if before_filtering - after_filtering:
                    LOG.warning(f"These pods will be ignored because of wrong NodeAffinity: "
                                f"{before_filtering - after_filtering}")

            # constant
            excluded_pods = ['elasticsearch-curator',
                             # SI pods
                             'ui-e2e-test', 'iam-bdd', 'conformance',
                             'test-stacklight', 'kaas-os-provider-e2e',
                             # to be checked
                             'patroni-lbauth',
                             'patroni-12-endpoint-cleanup',
                             'neutron-netns-cleanup-cron-default',
                             'ucp-secureoverlay',
                             'tf-test-tungsten-pytest',  # exist after TF test
                             'tf-vrouter-agent',  # checked as a part of tungsten_fabric.py:84
                             'tf-vrouter-vgw'  # checked as a part of tungsten_fabric.py:94
                             ]
            excluded_pods += self.EXCLUDED_PODS
            # special exclusions
            for excl_pod in excluded_pods:
                actual_list = \
                    [x for x in actual_list
                     if not x['metadata']['name'].startswith(excl_pod)]

            other_ns_pods = [pod for pod in actual_list
                             if pod['metadata']['namespace'] not in namespaces]
            actual_list = [pod for pod in actual_list
                           if pod['metadata']['namespace'] in namespaces]

            not_checked = [pod['metadata']['name'] for pod in actual_list]

            # sort names of expected pods so that
            # long names come first. "/no_owner" tag is excluded
            for ns in namespaces:
                expected_pods_lst = sorted(
                    list(expected_pods_dict[ns].keys()),
                    key=lambda x: len(x.split("/")[0]), reverse=True
                )
                for pod in expected_pods_lst:
                    num = expected_pods_dict[ns][pod]
                    prefix = pod.split("/")[0]
                    compare_list = \
                        [pod for pod in actual_list
                         if pod['metadata']['name'].startswith(prefix) and
                         pod['metadata']['name'] in not_checked]

                    if not compare_list:
                        failed[prefix] = {"actual": 0,
                                          "desired/expected": num}
                        continue

                    if '/no_owner' in pod:
                        # for pods that are marked "no_owner" we do not fetch
                        # owner and use # of pods from the file
                        LOG.debug(f"Number of pods "
                                  f"for {prefix} group will be checked "
                                  f"according to expected pods list")
                        desired_num = num
                    else:
                        # get owner kind and name for the first pod in list
                        first_pod = compare_list[0]
                        owner_references = \
                            first_pod['metadata']['owner_references']
                        kind_name = f"{owner_references[0]['kind']}/" \
                                    f"{owner_references[0]['name']}"

                        if kind_name.split('/')[0] == "Node":
                            # Starting from 1.17 mirror pods have
                            # Node/node-name in owner_references
                            LOG.warning(f"No replica count info for pod "
                                        f"{first_pod['metadata']['name']}. "
                                        f"Using expected number ({num}) "
                                        f"from the list")
                            # Better to handle this type of pods as '/no_owner'
                            LOG.info("Please add '/no_owner' for this group of"
                                     " pods to expected pods list file")
                            desired_num = num
                        else:
                            try:
                                desired_num = self.__get_replica_number(
                                    kind_name,
                                    first_pod['metadata']['namespace']
                                )
                                LOG.debug(f"First pod is "
                                          f"{first_pod['metadata']['name']}, "
                                          f"owner: {kind_name}, "
                                          f"replica #: {desired_num}")
                                if int(desired_num) != int(num):
                                    LOG.warning(f"Desired num ({desired_num}) "
                                                f"is not equal to the "
                                                f"number ({num}) in expected "
                                                f"pod list. Pod: {prefix}")
                            except (Exception, ApiException) as e:
                                LOG.error(e)
                                LOG.error(f"Cannot process {prefix} "
                                          f"group of pods. Skipping")
                                failed[prefix] = {"actual": 0,
                                                  "desired/expected": num}
                                continue
                    if int(desired_num) != len(compare_list):
                        failed[prefix] = {"actual": len(compare_list),
                                          "desired/expected": desired_num}
                    not_checked = [x for x in not_checked
                                   if x not in [pod['metadata']['name']
                                                for pod in compare_list]]

                    actual_ns = set([pod['metadata']['namespace']
                                     for pod in compare_list])
                    if set([ns]) != actual_ns:
                        failed[prefix] = {"actual namespace": actual_ns,
                                          "desired/expected": ns}

            if other_ns_pods:
                for pod in other_ns_pods:
                    LOG.error(f"Extra pod {pod['metadata']['name']} "
                              f"found in {pod['metadata']['namespace']} "
                              f"namespace")
                    if check_all_nss:
                        not_checked.append(pod['metadata']['name'])

            if failed or not_checked:
                result = {"Pods mismatch": failed,
                          "Not checked pods": not_checked}
                LOG.warning(f"Compare pod check failed: {result}")
                return result

        if not expected_pods:
            if exclude_pods:
                expected_pods, do = self.cluster.get_expected_objects(
                    exclude_pods=exclude_pods)
            else:
                expected_pods = self.cluster.expected_pods

        try:
            waiters.wait(lambda: not compare(
                expected_pods_dict=expected_pods,
                exclude_jobs=exclude_jobs),
                timeout=timeout, interval=interval)
        except exceptions.TimeoutError:
            result = compare(expected_pods_dict=expected_pods,
                             exclude_jobs=exclude_jobs)
            if result:
                err = f"Timeout waiting for pods. " \
                      f"After {timeout}s there are some fails: " \
                      f"{result}"
                raise TimeoutError(err)
        LOG.info("All pods and their replicas are found")

    @retry(AssertionError, delay=60, tries=10, logger=LOG)
    def check_k8s_resources_after_update(self, cluster_client,
                                         cluster_resources_before,
                                         reboot_required=False):
        # temporary exclude capacity/pods, check PRODX-11503
        # temporary exclude capacity/memory, check PRODX-18124
        # temporarily exclude feature.node.kubernetes.io/system-os_release*, check FIELD-5091
        # temporarily exclude system-os_release.ID
        # FIELD-5091 is fixed so OS is reported correctly and values are different from MKE 3.5.x
        labels_to_skip = {'capacity/pods',
                          'capacity/memory',
                          'labels/feature.node.kubernetes.io/system-os_release.VERSION_ID',
                          'labels/feature.node.kubernetes.io/system-os_release.VERSION_ID.major',
                          'labels/feature.node.kubernetes.io/system-os_release.VERSION_ID.minor',
                          'labels/feature.node.kubernetes.io/system-os_release.ID'}

        # exclude CPUID.SSE42 and CPUID.SSE4 labels as they may disappear during K8s upgrade
        # Also exclude iommu-enabled as it was deprecated in k8s: PRODX-30480
        gone_keys_to_skip = {'labels/feature.node.kubernetes.io/cpu-cpuid.SSE42',
                             'labels/feature.node.kubernetes.io/cpu-cpuid.SSE4',
                             'labels/feature.node.kubernetes.io/system-os_release.VERSION_ID',
                             'labels/feature.node.kubernetes.io/system-os_release.VERSION_ID.major',
                             'labels/feature.node.kubernetes.io/system-os_release.VERSION_ID.minor',
                             'labels/feature.node.kubernetes.io/system-os_release.ID',
                             'labels/feature.node.kubernetes.io/iommu-enabled'}
        new_keys_to_skip = {'labels/feature.node.kubernetes.io/cpu-cpuid.CMPXCHG8',
                            'labels/feature.node.kubernetes.io/cpu-cpuid.FXSR',
                            'labels/feature.node.kubernetes.io/cpu-cpuid.FXSROPT',
                            'labels/feature.node.kubernetes.io/cpu-cpuid.LAHF',
                            'labels/feature.node.kubernetes.io/cpu-cpuid.MOVBE',
                            'labels/feature.node.kubernetes.io/cpu-cpuid.OSXSAVE',
                            'labels/feature.node.kubernetes.io/cpu-cpuid.SCE',
                            'labels/feature.node.kubernetes.io/cpu-cpuid.X87',
                            'labels/feature.node.kubernetes.io/cpu-cpuid.XSAVE',
                            'labels/feature.node.kubernetes.io/cpu-hardware_multithreading',
                            'labels/feature.node.kubernetes.io/cpu-model.family',
                            'labels/feature.node.kubernetes.io/cpu-model.id',
                            'labels/feature.node.kubernetes.io/cpu-model.vendor_id',
                            'labels/feature.node.kubernetes.io/kernel-config.NO_HZ',
                            'labels/feature.node.kubernetes.io/kernel-config.NO_HZ_FULL',
                            'labels/feature.node.kubernetes.io/kernel-selinux.enabled'}

        if reboot_required:
            _reboot_skip_list = {'labels/feature.node.kubernetes.io/kernel-version.full',
                                 'labels/feature.node.kubernetes.io/kernel-version.major',
                                 'labels/feature.node.kubernetes.io/kernel-version.minor'}
            labels_to_skip.update(_reboot_skip_list)

        LOG.info(f'Labels to skip: {labels_to_skip}')
        LOG.info("Collect cluster fixed resources after update")
        cluster_resources_after = cluster_client.describe_fixed_resources()

        LOG.debug(f"Cluster resources before update:\n"
                  f"{yaml.dump(cluster_resources_before)}")
        LOG.debug(f"Cluster resources after update:\n"
                  f"{yaml.dump(cluster_resources_after)}")

        before_keys = cluster_resources_before.keys()
        after_keys = cluster_resources_after.keys()

        common_keys = before_keys & after_keys
        new_keys = after_keys - before_keys
        gone_keys = before_keys - after_keys

        LOG.info("Check for NEW appeared objects")
        new_objects = {}
        for k in new_keys:
            if not any([re.compile('node/.*/' + skip_key).match(k) for skip_key in new_keys_to_skip]):
                new_objects[k] = cluster_resources_after[k]
        if new_objects:
            LOG.info(f"The following NEW objects appeared:\n{yaml.dump(new_objects)}")

        LOG.info("Check for OLD disappeared objects")
        gone_objects = {}
        for k in gone_keys:
            if not any([re.compile('node/.*/' + skip_key).match(k) for skip_key in gone_keys_to_skip]):
                gone_objects[k] = cluster_resources_before[k]

        LOG.info("Check that fixed resources were not changed after update")
        before = {}
        after = {}
        for common_label in common_keys:
            if not any([skip_label in common_label for skip_label in labels_to_skip]):
                before[common_label] = cluster_resources_before[common_label]
                after[common_label] = cluster_resources_after[common_label]

        before = yaml.dump(before)
        after = yaml.dump(after)

        assert before == after, (f"Some cluster resources were changed."
                                 f"\nBefore:\n{before}\n"
                                 f"\nAfter:\n{after}")
        assert not gone_objects, (f"The following OLD objects disappeared:"
                                  f"\n{yaml.dump(gone_objects)}")

    def get_ucp_worker_agent_name(self):
        """Find the service name for 'ucp-worker-agent-*' for the current MKE
        """
        def filter_ucp_worker_agent_service(service):
            service_name = service['Spec']['Name']
            return (
                service_name.startswith('ucp-worker-agent-') and
                'ucp-worker-agent-win' not in service_name)

        dashboardclient = self.cluster.mkedashboardclient
        docker_services = dashboardclient.get_swarm_services()

        filtered_services = list(filter(filter_ucp_worker_agent_service,
                                        docker_services))
        if len(filtered_services) < 1:
            raise Exception(f"Docker service 'ucp-worker-agent-*' not found "
                            f"on nodes of the cluster {self._cluster.name}")

        service_names = {
            s['Spec']['Labels']['com.docker.ucp.version']: s['Spec']['Name']
            for s in filtered_services
        }
        LOG.debug(f"Found the following worker agents:\n{service_names}")
        latest_ucp_version = max(service_names.keys())
        service_name = service_names[latest_ucp_version]
        LOG.info(f"Current name of ucp-worker-agent-* service for MKE version "
                 f"{latest_ucp_version} is: {service_name}")
        return service_name

    def check_actual_expected_docker_services(self,
                                              expected_services=None,
                                              excluded_services=None,
                                              changed_after_upd=None,
                                              timeout=1200,
                                              interval=10):
        """Compare expected list of docker services (which are fetched
           automatically, unless explicitly provided) with actual
           list of docker services in this UCP cluster"""

        # storage for the status message from the latest check
        latest = {"status": ""}

        def compare(expected_services, excluded_services):
            """Get actual/expected docker service replicas and
               compare it to the expected_services from tests
               or from si_tests/templates/expected-pods/<cluserrelease>.yaml

               :rtype bool: False means that the errors were found
                            True means the check passed
            """
            failed = {}
            service_replicas = {}

            try:
                dashboardclient = self.cluster.mkedashboardclient
                docker_services = dashboardclient.get_swarm_service_replicas()
            except exceptions.TimeoutError as e:
                LOG.warning(f"Got a TimeoutError from compare_twice():\n{e}")
                raise e
            except Exception as e:
                msg = f"Error reading docker swarm services:\n{e}"
                LOG.warning(msg)
                latest["status"] = msg
                self.cluster.docker_ps()
                return False

            for name, replicas in docker_services.items():
                LOG.debug(f"Docker service: {name}: {replicas}")
                r_actual = int(replicas["RunningTasks"])
                r_expected = int(replicas["DesiredTasks"])
                if r_expected == 0:
                    LOG.debug(f"Skip docker service '{name}' "
                              f"because in docker defined 0 replicas")
                    continue
                if name in excluded_services:
                    LOG.debug(f"Skip docker service '{name}' "
                              f"because it is in the skip list")
                    continue
                service_replicas[name] = {}
                service_replicas[name]["actual"] = r_actual
                service_replicas[name]["expected"] = r_expected

            not_checked = service_replicas.keys()

            for svc in expected_services.keys():
                if svc in excluded_services:
                    LOG.debug(f"Skip expected service '{svc}' "
                              f"because it is in the skip list")
                    continue

                num = expected_services[svc]

                if svc not in service_replicas.keys():
                    failed[svc] = {"actual": 0,
                                   "test expected": num}
                    LOG.debug(f"Docker service '{svc}' not found "
                              f"while it is expected with '{num}' replicas")
                    continue

                r_actual = int(service_replicas[svc]["actual"])
                r_expected = int(service_replicas[svc]["expected"])
                if int(num) != r_expected or r_expected != r_actual:
                    failed[svc] = {"actual": r_actual,
                                   "docker expected": r_expected,
                                   "test expected": num}
                    LOG.debug(f"Docker service '{svc}' has wrong number "
                              f"of replicas: {failed[svc]}")

                not_checked = [x for x in not_checked if x != svc]

            if failed or not_checked:
                result = {"Docker services mismatch": failed,
                          "Not checked docker services": not_checked}
                LOG.warning(f"Compare docker services check failed: {result}")
                latest["status"] = result
                return False
            else:
                latest["status"] = ""
                LOG.info("Compare docker services check passed "
                         "(should not fail for 5 minutes)")
                return True

        def compare_twice(expected_services, excluded_services):
            """Ensure service statuses several times during 3 minutes

            If compare() method is passed - then repeat it, until
            timeout reached.
            TimeoutError means that the compare() method was
            never failed during the specified timeout, so return True
            """
            try:
                waiters.wait(lambda: not compare(
                    expected_services, excluded_services),
                    timeout=300, interval=20)
            except exceptions.TimeoutError:
                return True
            return False

        if not expected_services:
            docker_objects = self.cluster.expected_docker_objects
            expected_services = docker_objects["service"]
            if changed_after_upd:
                # proceed with a static copy of keys, to not replace keys twice
                expected_services_keys = list(expected_services.keys())
                for key in changed_after_upd:
                    if key in expected_services_keys:
                        key_data = expected_services[key]
                        del expected_services[key]
                        expected_services[changed_after_upd[key]] = key_data

        # list of excluded docker services
        excluded_services = excluded_services or []
        LOG.info(f"Skip checks for the following docker services: "
                 f"{excluded_services}")

        try:
            waiters.wait(lambda: compare_twice(
                expected_services, excluded_services),
                timeout=timeout, interval=interval)
        except exceptions.TimeoutError:
            if latest['status']:
                err = (f"Timeout waiting for docker services. "
                       f"After {timeout}s there are some fails: "
                       f"{latest['status']}")
            else:
                err = "Timeout waiting for docker services"
            raise TimeoutError(err)
        LOG.info("All docker services and their replicas were found")

    def __get_replica_number(self, obj, namespace):
        """Return desired number of replicas/scheduled pods according to
           the pod owner specification (e.g. statefulset).
        """
        kind, name = obj.split('/')
        desired_num = 0
        if kind == 'ReplicaSet':
            rs = self.k8sclient.replicasets.get(
                name=name, namespace=namespace).read().to_dict()
            desired_num = rs['metadata']['annotations'][
                'deployment.kubernetes.io/desired-replicas']
        elif kind == 'StatefulSet':
            ss = self.k8sclient.api_apps.read_namespaced_stateful_set(
                name=name, namespace=namespace).to_dict()
            desired_num = ss['spec']['replicas']
        elif kind == 'DaemonSet':
            ds = self.k8sclient.daemonsets.get(
                name=name, namespace=namespace).read().to_dict()
            desired_num = ds['status']['desired_number_scheduled']
        else:
            LOG.warning(f"Unknown kind {kind}")
        return desired_num

    def check_actual_expected_replicas_count(self, replicas_list, expected_replicas_count=1):
        """
        replicas_list: List of replicaset objects
        """
        checked_replicas = []
        for r in replicas_list:
            r_status = r.data.get('status') or {}
            current_replicas_count = r_status.get('replicas', None)
            if current_replicas_count == expected_replicas_count:
                checked_replicas.append(r)
            else:
                LOG.warning(f'Replicas count for replicaset {r.name} not as expected. \n'
                            f'Current replicas {current_replicas_count}, expected: {expected_replicas_count}')
        if len(checked_replicas) == len(replicas_list):
            LOG.info("All replicasets has replicas count as expected")
            return True
        LOG.warning("Not all replicasets has expected replicas count")
        return False

    def check_deployments_ready(self, deployments_list):
        for d in deployments_list:
            if d.wait_ready():
                continue
            else:
                LOG.error(f'Deployment {d} is not ready')
                return False
        LOG.info(f'All deployments in: {deployments_list }have status ready')
        return True

    def check_cluster_deleted(self, timeout=1800, interval=15):
        waiters.wait(lambda: not self.cluster.is_existed(),
                     timeout=timeout, interval=interval,
                     timeout_msg='Timeout waiting for cluster deletion')
        LOG.info(f"Cluster {self.cluster.name} has been deleted")

    def _get_cluster_readiness(self,
                               exp_provider_status=True,
                               expected_fails=None,
                               expected_condition_fails=None):
        """Get conditions from Cluster and Machine objects

        :rtype bool: True only if all conditions are true, False in other case
        """
        machines_result = [m.are_conditions_ready(expected_condition_fails,
                                                  verbose=True)
                           for m in self.cluster.get_machines()]
        cluster_result = [
            self.cluster.are_conditions_ready(expected_condition_fails,
                                              verbose=True),
            self.cluster.is_ready(exp_provider_status, expected_fails)]

        return all(machines_result + cluster_result)

    def _get_cluster_status_msg(self):
        """To get latest status of the Cluster and Machine objects"""
        status = self.cluster.data['status']['providerStatus']
        status_keys = [
            'bastion',
            'conditions',
            'helm',
            'kind',
            'loadBalancerHost',
            'loadBalancerStatus',
            'nodes',
            'notReadyObjects',
            'ready',
        ]
        cluster_status = {k: v for k, v in status.items() if k in status_keys}
        machines = {m.name: (m.data.get('status') or {}).get('providerStatus', {})
                    for m in self.cluster.get_machines()}
        machines_msg = []
        for name, status in machines.items():
            prefix = f"Machine {self.cluster.namespace}/{name}"
            if 'conditions' not in status:
                msg = f"{prefix}: NO CONDITIONS IN PROVIDER STATUS"
                machines_msg.append(msg)
            else:
                errors = [f"{c['type']} condition: \"{c['message']}\""
                          for c in status['conditions'] if not c['ready']]
                if errors:
                    machines_msg.append(f"[ {prefix} ] " + "; ".join(errors))

        machines_str = "\n".join(machines_msg)
        return (f"Cluster data status:\n{yaml.dump(cluster_status)}\n"
                f"{machines_str}")

    def check_cluster_readiness(
            self,
            timeout=settings.CHECK_CLUSTER_READINESS_TIMEOUT,
            interval=60,
            exp_provider_status=True,
            expected_fails=None,
            expected_condition_fails=None):
        """
        Check that overall cluster is Ready
        and no unexpected deployments failed
        :param timeout: timeout for waiter
        :type timeout: int
        :param interval: interval to check status
        :type interval: int
        :param exp_provider_status: expected overall cluster status
        :type exp_provider_status: bool
        :param expected_fails: list of strings with
        expectedly failed deployment names
        :type expected_fails: List[str]
        :param expected_condition_fails: dict with conditions like
            { <condition type>: <part of condition message to match> , }
        :type expected_condition_fails: Dict[str]

        :rtype bool: bool
        """
        LOG.info("Checking cluster readiness")
        expected_fails = expected_fails or []
        cluster_release = self.cluster.clusterrelease_version
        LOG.info(f"Cluster release: {cluster_release}")
        can_check_mke = 'mke' in cluster_release and \
                        version.parse(cluster_release) \
                        > version.parse('mke-5-13-0-3-3-6')
        can_check_mos = 'mos' in cluster_release and \
                        version.parse(cluster_release) \
                        > version.parse('mos-6-12-0-21-1')

        if can_check_mke or can_check_mos:

            timeout_msg = "Timeout waiting for cluster readiness"
            try:
                waiters.wait(
                    lambda: self._get_cluster_readiness(
                        exp_provider_status,
                        expected_fails,
                        expected_condition_fails),
                    timeout=timeout,
                    interval=interval,
                    timeout_msg=timeout_msg,
                    status_msg_function=self._get_cluster_status_msg)
            except exceptions.TimeoutError as e:
                if ['Ceph'] == collect_cluster_readiness.readiness.get('not_ready', []):
                    waiters.wait(
                        lambda: self._get_cluster_readiness(
                            exp_provider_status,
                            expected_fails,
                            expected_condition_fails),
                        timeout=600,
                        interval=60,
                        timeout_msg=timeout_msg,
                        status_msg_function=self._get_cluster_status_msg)
                else:
                    # Raise exception if not only Ceph
                    raise e

            LOG.info("Cluster checked")
        else:
            LOG.info(f"Skip cluster readiness check for cluster release "
                     f"{cluster_release} as feature is not supported yet")

    def check_offline_isolation(self, proxy_access_str):
        resources_to_check = [
            # Binaries
            "https://binary.mirantis.com",
            # Docker images
            "http://mirantis.azurecr.io",
            # APT mirror
            "https://mirror.mirantis.com",
            # Docker Engine mirror
            "https://repos.mirantis.com",
        ]
        cert_cmd = ""
        certificate_param = ""

        if settings.KAAS_SSL_PROXY_CERTIFICATE_FILE:
            proxy_name = self.cluster.spec['providerSpec']['value']['proxy']
            if not self.cluster.is_management:
                proxy_obj = self.cluster.get_parent_cluster().k8sclient.kaas_proxies.get(
                    proxy_name, self.cluster.namespace)
            else:
                proxy_obj = self.cluster. k8sclient.kaas_proxies.get(proxy_name, self.cluster.namespace)
            proxy_cert_b64 = proxy_obj.data['spec']['caCertificate']
            cert_cmd = f'echo {proxy_cert_b64}|base64 -d > custom_proxy_cert;'
            certificate_param = "--cacert custom_proxy_cert"

        machines = self.cluster.get_machines()
        machines_to_check = {machine.machine_type: machine
                             for machine in machines}
        for machine_type, machine in machines_to_check.items():
            LOG.info(f"Offline isolation check on machine [{machine_type}] "
                     f"{self._cluster.name}/{machine.name}")

            # Negative check: direct access from the offline node
            # to external resources should fail.
            # * "hostname -s" is used because equinix labs has wrong name
            #   order in /etc/hosts: "hostname -f" returns "localhost"
            for resource in resources_to_check:
                negative_cmd = (
                    f"set -x; hostname -s;"
                    f"unset HTTP_PROXY; unset HTTPS_PROXY;"
                    f"unset http_proxy; unset https_proxy;"
                    f"curl --connect-timeout 10 -sS {resource}"
                )
                result = machine.exec_pod_cmd(negative_cmd, verbose=False)
                assert result['exit_code'] != 0, (
                    f"No offline isolation on the machine "
                    f"{self._cluster.name}/{machine.name}:"
                    f"\n{result['logs']}")
                assert (
                    "Connection timed out" in result['logs'] or
                    "Failed connect to" in result['logs'] or
                    "Failed to connect to" in result['logs'] or
                    "Could not resolve host" in result['logs'] or
                    "Resolving timed out" in result['logs']
                ), (
                    f"Unexpected result while checking offline isolation "
                    f"on the machine {self._cluster.name}/{machine.name}:"
                    f"\n{result['logs']}")

            if proxy_access_str:
                # Positive check: access from the offline node to
                # an external resource using the specified proxy should pass
                positive_cmd = (
                    f"set -x; hostname -s;"
                    f"{cert_cmd}"
                    f"curl --connect-timeout 10 -sS -x '{proxy_access_str}' "
                    f"{resources_to_check[0]} "
                    f"{certificate_param}"
                )
                result = machine.exec_pod_cmd(positive_cmd, verbose=False)
                assert result['exit_code'] == 0, (
                    f"No proxy access from the machine "
                    f"{self._cluster.name}/{machine.name}:"
                    f"\n{result['logs']}")
            else:
                LOG.warning("Skip external access check through proxy "
                            "because proxy setting is empty")

            LOG.info(f"Offline isolation check on machine [{machine_type}] "
                     f"{self._cluster.name}/{machine.name}: PASSED")

    @staticmethod
    @retry(AssertionError, delay=60, tries=10, logger=LOG)
    def check_cert_conversation(host, port=443):
        try:
            ssl_connection_setting = Context(SSLv23_METHOD)
        except ValueError:
            ssl_connection_setting = Context(TLSv1_2_METHOD)
        ssl_connection_setting.set_timeout(5)
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((host, port))
            c = Connection(ssl_connection_setting, s)
            c.set_tlsext_host_name(str.encode(host))
            c.set_connect_state()
            c.do_handshake()
            cert = c.get_peer_certificate()
            LOG.info(f'Certificate: {cert}')
            subject_list = cert.get_subject().get_components()
            LOG.info(f'Subject list: {subject_list}')
            cert_byte_arr_decoded = {}
            for item in subject_list:
                cert_byte_arr_decoded.update(
                    {item[0].decode('utf-8'): item[1].decode('utf-8')})
            LOG.info(f"Decoded cert byte array: {cert_byte_arr_decoded}")

            assert len(cert_byte_arr_decoded) > 0
            assert cert_byte_arr_decoded["CN"] == host, \
                'Cert common name is not equal to expected one'

    def check_cert_equal(self, port=443, cert=None):
        LOG.info("Check cert equal")
        cmd = f"openssl s_client -connect 127.0.0.1:{port} 2>/dev/null | openssl x509"
        machines = self.cluster.get_machines(machine_type='control')
        if not cert:
            cert = machines[0].exec_pod_cmd(cmd)['logs'][:-1]
        invalid_cert = []
        for m in machines:
            LOG.debug(f"{m.name}  - cert: ")
            local_cert = m.exec_pod_cmd(cmd)['logs']
            if local_cert[:-1] != cert:
                invalid_cert.append(m.name)
        assert not invalid_cert, f"Nodes with invalid cert: {invalid_cert}"

    @retry(AssertionError, delay=30, tries=10, logger=LOG)
    def validate_machines_mcc_ca(self, ca):
        machines = self.cluster.get_machines()
        mcc_ca_cmd = "sudo cat /usr/local/share/mcc/mcc.crt"

        for m in machines:
            mcc_ca_file = m.exec_pod_cmd(mcc_ca_cmd)['logs']
            assert ca in mcc_ca_file, f"Node with invalid mcc.crt: {m.name}"

    def check_actual_expected_ceph_pods(self, scale_machines_names_list=None,
                                        timeout=1200, interval=120):
        machines_to_count = []
        if self.cluster.provider not in utils.Provider.with_ceph():
            LOG.info(f"Provider '{self.cluster.provider}' doesn't support Ceph. "
                     f"Skipping Ceph pods checks")
            return True
        elif scale_machines_names_list and self.cluster.provider == utils.Provider.equinixmetal:
            # Check if node is used for ceph
            # If Ceph configuration is not manual, then all machines are used
            # for Ceph. Else, need to check this
            if self.cluster.data.get('spec', {}).get(
                'providerSpec', {}).get('value', {}).get('ceph', {}).get(
                    'manualConfiguration', False):
                for machine in scale_machines_names_list:
                    machine_obj = self.cluster.get_machine(machine)
                    storage = machine_obj.data.get(
                        'spec').get('providerSpec').get(
                        'value').get('ceph', {})
                    if storage.get('managerMonitor', False) \
                            or storage.get('storage', False):
                        machines_to_count.append(machine)
                    else:
                        LOG.info(f"Machine {machine} was not used for Ceph. "
                                 f"Skipping")
            else:
                LOG.info(f"All of scale machines {scale_machines_names_list} "
                         f"are used for Ceph, because "
                         f"Ceph configuration is not manual")
                machines_to_count.extend(scale_machines_names_list)
        elif scale_machines_names_list and self.cluster.provider == utils.Provider.baremetal:
            # Need to check is scale node is used for Ceph or not
            # For BM Provider this is done by labeling node as storage
            for machine in scale_machines_names_list:
                machine_obj = self.cluster.get_machine(machine)
                if 'storage' in machine_obj.data.get('metadata', {}).get(
                        'labels'):
                    machines_to_count.append(machine)

        def check_machines_in_cephcluster():
            is_machine_added = True
            ceph_machines =\
                self.cluster.get_kaascephcluster().data.get(
                    'spec', {}).get('cephClusterSpec', {}).get(
                    'nodes', {}).keys()
            for machine in machines_to_count:
                if machine in ceph_machines:
                    LOG.info(f"Machine {machine} was succesfully "
                             f"added to kaascephcluster")
                    continue
                else:
                    LOG.info(f"Machine {machine} is not existed "
                             f"in kaascephcluster yet")
                    is_machine_added = False
            return is_machine_added
        if machines_to_count:
            # Wait for machines are added to kaascephcluster
            waiters.wait(
                check_machines_in_cephcluster,
                timeout=timeout,
                interval=interval)

        def check_ceph_pods_number():
            expected_osd_pod_num = 0
            expected_mon_pod_num = 0
            if machines_to_count:
                # For equinix provider all of disks (except boot) are used
                # for Ceph if node is used for Ceph. So we need to check
                # this condition too. Check that free disks number is equal
                # storage devices in kaascephcluster after adding new node
                if self.cluster.provider in [utils.Provider.equinixmetal, utils.Provider.equinixmetalv2]:
                    disks_missmatch = {}
                    for machine in machines_to_count:
                        # Get free disks
                        machine_data = self.cluster.get_machine(
                            machine).data
                        free_disks = [d for d in machine_data.get(
                            'status', {}).get('providerStatus', {}).get(
                            'hardware', {}).get('storage', {}) if not d.get(
                            'isBoot', False)]
                        # Get Ceph disks from kaascephcluster
                        kaascephcluster_storage =\
                            self.cluster.get_kaascephcluster().data.get(
                                'spec', {}).get('cephClusterSpec', {}).get(
                                'nodes', {}).get(machine, {}).get(
                                'storageDevices', [])
                        # Check it matches
                        if len(free_disks) != len(kaascephcluster_storage):
                            disks_missmatch[machine] = {
                                'node_disks': free_disks,
                                'ceph_disks': kaascephcluster_storage}
                    if disks_missmatch:
                        raise Exception(
                            f"Disks and osds count missmatch "
                            f"for next nodes: {disks_missmatch.keys()}."
                            f"\n{yaml.dump(disks_missmatch)}")

            missmatch = {}
            ceph_cluster = self.cluster.get_kaascephcluster()
            ceph_nodes = ceph_cluster.data.get(
                'spec', {}).get('cephClusterSpec', {}).get('nodes', {})
            for k, v in ceph_nodes.items():
                stordev = [dev.get('name') for dev in v.get(
                    'storageDevices', [])]
                expected_osd_pod_num += len(stordev)
                if 'mon' in v.get('roles', []):
                    expected_mon_pod_num += 1
            actual_osd_pods = [
                pod for pod in self.cluster.k8sclient.pods.list_starts_with(
                    pattern='rook-ceph-osd-', namespace='rook-ceph') if
                'osd-prepare' not in pod.name]
            actual_mon_pods = [
                pod for pod in self.cluster.k8sclient.pods.list_starts_with(
                    pattern='rook-ceph-mon-', namespace='rook-ceph')]
            actual_osd_pods_num = len(actual_osd_pods)
            actual_mon_pods_num = len(actual_mon_pods)
            mssg = "Missmatch between expected and actual Ceph pods number."
            if actual_osd_pods_num != expected_osd_pod_num:
                mssg += (f" Current osds count {actual_osd_pods_num}, "
                         f"but should be {expected_osd_pod_num}.")
                missmatch['osd'] = actual_osd_pods_num

                if actual_mon_pods_num != expected_mon_pod_num:
                    mssg += (f" Current mons count {actual_mon_pods_num}, "
                             f"but should be {expected_mon_pod_num}.")
                    missmatch['mon'] = actual_mon_pods_num
            if missmatch:
                LOG.error(mssg)
                return False
            else:
                LOG.info("Current osds and mons pods number as expected")
                return True

        waiters.wait(lambda: check_ceph_pods_number(), timeout=timeout,
                     interval=interval)

    def check_pod_container_expected_resources(
            self, pod_name, pod_ns, pod_container_name=None,
            expected_resources=None):
        """
        Compare existed pod resources and expected.
        expected_resources should be in next format:
          limits:
            cpu: "3"
            memory: 2Gi
          requests:
            cpu: "2"
            memory: 1Gi

        Use None if you expect no resources. For example:
          limits: None
          requests: None
        """

        pod = self.cluster.k8sclient.pods.get(
            name=pod_name, namespace=pod_ns)
        pod_name = pod.name
        if not pod_container_name:
            assert len(pod.containers) == 1, (
                f"Container name is not set and there are multiple containers "
                f"in pod {pod_name}. Please, choose one of "
                f"{[c['name'] for c in pod.containers]}")
            pod_container_name = pod.containers[0]['name']

        if not expected_resources:
            LOG.info(f"Expected resources are not set for pod {pod_name}. "
                     f"Skipping checks. "
                     f"Current resources: "
                     f"\n{yaml.dump(pod.containers_resources)}")
            return True

        resources = pod.containers_resources
        expected_limits = expected_resources.get('limits', {})
        expected_requests = expected_resources.get('requests', {})
        container_data = [c for c in resources['containers'] if
                          c['name'] == pod_container_name]
        assert container_data, (f"Container {pod_container_name} not found "
                                f"for pod {pod_name}")
        container = container_data[0]
        container_limits = container.get('resources').get('limits')
        container_requests = container.get('resources').get('requests')
        if expected_limits == container_limits:
            if expected_requests == container_requests:
                LOG.info(f"All resources for container {pod_container_name} "
                         f"in pod {pod_name} as expected")
                return True
            else:
                LOG.error(f"Limits as expected, but requests are not as "
                          f"expected for container {pod_container_name} "
                          f"in pod {pod_name}")
                LOG.debug(f"\nExpected resources: "
                          f"\n{yaml.dump(expected_resources)} "
                          f"\nCurrent resources: "
                          f"\n{yaml.dump(resources)}")
        else:
            if expected_requests != container_requests:
                LOG.error(f"Limits and requests are not as expected "
                          f"for container {pod_container_name} in pod "
                          f"{pod_name}")
                LOG.debug(f"\nExpected resources: "
                          f"\n{yaml.dump(expected_resources)} "
                          f"\nCurrent resources: "
                          f"\n{yaml.dump(resources)}")
            else:
                LOG.error(f"Limits are not as expected for container "
                          f"{pod_container_name} in pod {pod_name}")
                LOG.debug(f"\nExpected resources: "
                          f"\n{yaml.dump(expected_resources)} "
                          f"\nCurrent resources: "
                          f"\n{yaml.dump(resources)}")
        return False

    def compare_supported_and_existed_clusterreleases(self):
        if not self.cluster.is_management:
            LOG.info("KaasRelease and ClusterRelease objects available "
                     "only for mgmt cluster")
            return None
        lost_releases = []
        clusterreleases_objects = \
            self.cluster.k8sclient.kaas_clusterreleases.list_all()
        clusterreleases_objects_names = [
            cl_r.name for cl_r in clusterreleases_objects]
        LOG.debug(f"Clusterreleases objects names: "
                  f"{clusterreleases_objects_names}")
        supported_clusterreleases = \
            self.cluster._manager.get_supported_clusterreleases()
        supported_clusterreleases_names = \
            [v.get('name', '') for k, v in supported_clusterreleases.items()]
        LOG.debug(f"Supported clusterreleases names: "
                  f"{supported_clusterreleases_names}")
        for name in supported_clusterreleases_names:
            if name not in clusterreleases_objects_names:
                lost_releases.append(name)
        if lost_releases:
            raise Exception(f"Next releases are announced as Supported, but "
                            f"clusterrelease objects for these releases are "
                            f"not existed: {lost_releases}")

    def check_actual_expected_rolebindings(self, timeout=300, interval=10):
        # Check posisble on cluster with mke-7-2-0 and larger
        cluster_release = self.cluster.clusterrelease_version
        LOG.info(f"Cluster release: {cluster_release}")
        can_check_mke = 'mke' in cluster_release and \
                        version.parse(cluster_release) \
                        > version.parse('mke-7-2-0-3-4-0')

        def compare(expected_rolebinds_dict: dict):
            failed = {}
            namespaces = expected_rolebinds_dict.keys()
            actual_list = [rolebind for rolebind in
                           self.k8sclient.rolebindings.list_raw().to_dict()['items']]

            # constant
            excluded_rolesbinds = []
            excluded_rolesbinds += self.EXCLUDED_ROLESBINDS
            # special exclusions
            for one_excl in excluded_rolesbinds:
                actual_list = [x for x in actual_list
                               if not x['metadata']['name'].startswith(one_excl)]

            actual_list = [rolebind for rolebind in actual_list
                           if rolebind['metadata']['namespace'] in namespaces]

            not_checked = [(rolebind['metadata']['name'], rolebind['metadata']['namespace'])
                           for rolebind in actual_list]

            for ns in namespaces:
                expected_rolebinds_lst = sorted(
                    list(expected_rolebinds_dict[ns].keys()), reverse=True
                )
                for rolebind in expected_rolebinds_lst:
                    roleref = expected_rolebinds_dict[ns][rolebind]
                    compare_rolebind = next((rb for rb in actual_list if
                                             rb['metadata']['name'].startswith(rolebind) and
                                             rb['metadata']['namespace'] == ns and
                                             (rb['metadata']['name'], rb['metadata']['namespace']) in not_checked),
                                            None)

                    if not compare_rolebind:
                        failed[rolebind] = {"actual": None,
                                            "desired/expected": roleref,
                                            "namespace": ns}
                        continue

                    # desired_kind, desired_rolename = roleref.split('/')
                    compare_rolebind_roleref = compare_rolebind['role_ref']
                    compare_rolebind_roleref = (f"{compare_rolebind_roleref['kind']}/"
                                                f"{compare_rolebind_roleref['name']}")
                    if roleref != compare_rolebind_roleref:
                        failed[rolebind] = {"actual": compare_rolebind_roleref,
                                            "desired/expected": roleref,
                                            "namespace": ns}

                    not_checked = [x for x in not_checked
                                   if x != (rolebind, ns)]

            if failed or not_checked:
                result = {"RoleBind mismatch": failed,
                          "Not checked rolebinds": not_checked}
                LOG.warning(f"Compare rolebindings check failed: {result}")
                return result

        if can_check_mke:
            expected_rolebindings = self.cluster.get_expected_rolebindings()

            try:
                waiters.wait(lambda: not compare(
                    expected_rolebinds_dict=expected_rolebindings),
                    timeout=timeout, interval=interval)
            except exceptions.TimeoutError:
                result = compare(expected_rolebinds_dict=expected_rolebindings)
                if result:
                    err = f"Timeout waiting for rolebindings. " \
                        f"After {timeout}s there are some fails: " \
                        f"{result}"
                    raise TimeoutError(err)
            LOG.info("All expected rolebindings are placed")
        else:
            LOG.info(f"Skip cluster rolebindings check for cluster release "
                     f"{cluster_release} as feature is not supported yet")

    def get_machine_repository_url(self, machine):
        """
        Get /etc/apt/sources.list for machine
        Args:
            machine: machine class

        Returns: output for the executed command

        """
        output = machine.exec_pod_cmd("grep deb /etc/apt/sources.list", verbose=False)['logs']
        return output

    def check_repository_url(self):
        """
        Check repository url for each machine

        Returns: None or Exception

        """
        def _check_url(machines, machine_type):
            """
            Check if repository url (separate for each machine type) contains expected url for each machine.
            Args:
                machines: list of machine objects
                machine_type: machine type

            Returns: list of machines that not match

            """
            not_match = []

            for machine in machines:
                LOG.info(f"Get repository url for {machine_type} machine {machine.name}")
                output = self.get_machine_repository_url(machine)
                result = re.findall(r"https://[^\s]*", output)
                # check all except the last one as it relates to kubernetes-extra repo
                if not all(line.endswith(repo_url[machine_type]) for line in result[:-1]):
                    not_match.append(machine.name)
                    LOG.info(
                        f"Expected repository url is '{repo_url[machine_type]}' "
                        f"but machine {machine.name} has:\n {output}")
            return not_match

        repo_url = self.cluster.get_cluster_repository_url()
        control_machines = self.cluster.get_machines(machine_type='control')
        worker_machines = self.cluster.get_machines(machine_type='worker')

        errors = _check_url(control_machines, 'control') + _check_url(worker_machines, 'worker')

        if errors:
            msg = f"Next machine has incorrect repository url: {', '.join(errors)}"
            raise Exception(msg)

    def expect_kernel_changed_between_versions(self, cr_previous=None):
        """Compare expected kernel versions with previous and current clusterreleases,
        for case, where we dont have allowedDistributions """
        kernel_changed = False
        expected_versions_map_previous = self.cluster.get_expected_kernel_version(target_cr=cr_previous)
        expected_versions_map_current = self.cluster.get_expected_kernel_version()
        for k, v in expected_versions_map_current.items():
            if expected_versions_map_previous.get(k) is not None:
                if expected_versions_map_previous[k] != expected_versions_map_current[k]:
                    expected_versions_map_previous_formatted = [
                        el for el in expected_versions_map_previous[k] if 'notgreenfield' not in el]
                    expected_versions_map_current_formatted = [
                        el for el in expected_versions_map_current[k] if 'notgreenfield' not in el]
                    if expected_versions_map_previous_formatted != expected_versions_map_current_formatted:
                        LOG.debug(f"Expected version current: {expected_versions_map_current_formatted}")
                        LOG.info(f"Expected version previous {expected_versions_map_previous_formatted}")
                        kernel_changed = True
                        LOG.info(f"Kernels should be changed between upgrade "
                                 f"({cr_previous} -> {self.cluster.clusterrelease_version}) "
                                 f"from {expected_versions_map_previous_formatted} "
                                 f"to {expected_versions_map_current_formatted} for {k}")

        return kernel_changed

    def check_actual_expected_kernel_versions_simple(self, expected_versions_map):
        """Compare expected kernel version with actual, per machine,
        for case, where we dont have allowedDistributions """
        # TODO(alexz): Depreacted, remove after 2.17 release
        actual_os, actual_kernel = self.cluster.get_nodes_kernel_versions()
        for k, v in actual_kernel.items():
            a_log = f"{actual_os[k]} {actual_kernel[k]}"
            assert v in expected_versions_map.get(actual_os[k], ''), \
                f"Mismatch between kernel version on {k}. " \
                f"actual: {a_log}, expected: " \
                f"{json.dumps(expected_versions_map)}"
        LOG.info("Kernel versions are valid on all nodes")
        return

    def check_actual_expected_kernel_versions(self):
        """
        Compare expected kernel version with actual, per machine
        TODO(alexz): part with "allowedDistributions" still on WIP phase, and upcoming changes will be
        integrated later. New logic will not affect old one style, due hide under FF.
        """
        expected_versions_map = self.cluster.get_expected_kernel_version()
        kaas_version = self.cluster.get_kaasrelease_version()
        # SI-tests now will be able to work only with 2.16+ releases
        assert "allowedDistributions" in expected_versions_map.keys(), \
            "allowedDistributions not found in expected_versions_map from test data."

        LOG.warning("'allowedDistributions' extra mapping detected, per-machine logic will be used")
        test_allowed_dists = expected_versions_map['allowedDistributions']
        #
        kaas_release_name = self.cluster.clusterrelease_version
        kaas_release_allowed_distr = self._cluster._manager.get_clusterrelease(name=kaas_release_name
                                                                               ).data['spec']['allowedDistributions']
        kaas_release_default_dist = [d for d in kaas_release_allowed_distr if d.get('default')]
        assert len(kaas_release_default_dist) == 1, \
            f"Expected to have only one default distib in release!ClusterRelease-map:" \
            f"\n{kaas_release_allowed_distr}\n is broken!"

        LOG.info(f"ClusterRelease:{kaas_release_name} have allowedDistributions:\n{kaas_release_allowed_distr}")
        perhostdata = self.cluster.get_nodes_kernel_and_tarfs_versions()
        # TODO(alexz): previous information required to determine initial tarfs. Currently its not possible to do.
        # corresponding changes will be integrated into lcm-agent(?) and test will be refactored accordingly.
        # Dummy check, is cluster been upgraded
        is_upgraded = False
        previous_clusterrelease = self.cluster.data['status'].get('providerStatus', {}).get(
            'releaseRefs', {}).get('previous', {}).get('name', {})
        if previous_clusterrelease:
            LOG.warning("Upgrade has been detected. Test-check for correct version of initial tarfs(dib_datetime)"
                        "has been disabled - now, we check only that node in AllowedVersions for release")
            is_upgraded = True
        # as for now, we may check only fact that dist in allowed, not exact distr version ot be deployed
        # for example - if we have multiply `ubuntu/focal` - we cant check that it was deployed from EXACT id.
        test_allowed_distr_ids = [d['id'] for d in test_allowed_dists]

        for m in self.cluster.get_machines():
            kaas_exp_distr_id = m.data['spec']['providerSpec']['value'].get("distribution", None)

            if not kaas_exp_distr_id and version.parse(kaas_version) <= version.parse("kaas-2-16-1"):
                fallback_distr = 'ubuntu/bionic'
                # todo: remove after 2.18 release.
                #  Transition period, it is still allowed to have fallback distr, which will be
                #  automatically set to bionic during upgrade to 2.17.
                LOG.warning(f'Fallback to default distribution: {fallback_distr} ')
                kaas_exp_distr_id = fallback_distr

            if kaas_exp_distr_id is None:
                LOG.error(f"Exact 'distribution' reference not found in spec machine:{m.name}\n"
                          f"This is unexpected for release testing. Must be one of {test_allowed_distr_ids}")
                raise Exception(f'Unexpected distribution spec {kaas_exp_distr_id}')
            # check for upgraded case
            # TODO(alexz): looks like there is no way in kaas, to identify machine initial deployment version.
            # so for upgrade check, we need to disable those check.
            # here we check that node has been initially provisioned from expected tarfs.
            # Those info are always initial, even after upgrade\reboot.
            if is_upgraded:
                LOG.debug("TODO(alexz): here(?) should be actual check for repository change after upgrade")
                test_allowed_distr_kernels = [d['kernel'] for d in test_allowed_dists if d['id'] == kaas_exp_distr_id]
                test_allowed_distr_os_vers = [d['os_version'] for d in test_allowed_dists
                                              if d['id'] == kaas_exp_distr_id]
                assert perhostdata[m.name]['kernel'] in test_allowed_distr_kernels, \
                    f"Expected machine=kernel on host os: {m.name} in {test_allowed_distr_kernels}, but have:\n" \
                    f"{perhostdata[m.name]['kernel']}"
                assert perhostdata[m.name]['os_version'] in test_allowed_distr_os_vers, \
                    f"Expected machine=os_version on host os: {m.name} in {test_allowed_distr_os_vers}, but have:\n" \
                    f"{perhostdata[m.name]['os_version']}"
                LOG.warning(f"machine dib_version check skipped due upgrade. "
                            f"Detected on node currently:{perhostdata[m.name]['dib_datetime']}")
                # after si-tests check finished, we may check expected reports from lcm-agent,
                # about actial information from node, but in perspective of lcm-agent. PRODX-22958 might be related.
                host_info = \
                    self.cluster.get_cluster_lcmmachine(name=m.name,
                                                        namespace=self.cluster.namespace).data['status']['hostInfo']
                assert host_info['kernelVersion'] in test_allowed_distr_kernels, \
                    f"Expected machine=kernel on host os: {m.name} in {test_allowed_distr_kernels}," \
                    f"but lcm-agent report:\n {host_info['kernelVersion']}"
                return
            # check for initial deployment, exact check for data 1:1, w\o upgrade
            else:
                test_data_for_exp_dist = [d for d in test_allowed_dists if
                                          d['id'] == kaas_exp_distr_id and not d.get('notgreenfield', False)]
                # that logic not exists yet in core itself, also tests should be in "is_upgraded" switch.
                assert len(test_data_for_exp_dist) == 1, \
                    "Multiply distribs found. Please update test logic, according to new logic for " \
                    "upgrade-path selector in kaas/core."
                LOG.info(f"For machine:{m.name}:\n"
                         f"KaaS deployed OS-id:{kaas_exp_distr_id}\n"
                         f"Test expected data for distr:{test_data_for_exp_dist}")

                test_allowed_distr_kernel = test_data_for_exp_dist[0]['kernel']
                test_allowed_distr_os_ver = test_data_for_exp_dist[0]['os_version']
                test_allowed_distr_dib_ver = test_data_for_exp_dist[0]['dib_version']
                assert perhostdata[m.name]['kernel'] == test_allowed_distr_kernel, \
                    f"Expected machine=kernel on host os: {m.name} == {test_allowed_distr_kernel}, but have:\n" \
                    f"{perhostdata[m.name]['kernel']}"
                assert perhostdata[m.name]['os_version'] == test_allowed_distr_os_ver, \
                    f"Expected machine=os_version on host os: {m.name} == {test_allowed_distr_os_ver}, but have:\n" \
                    f"{perhostdata[m.name]['os_version']}"

                assert perhostdata[m.name]['dib_datetime'] == test_allowed_distr_dib_ver, \
                    f"Expected machine=dib_version on host os: {m.name} == {test_allowed_distr_dib_ver}, but have:\n" \
                    f"{perhostdata[m.name]['dib_datetime']}"
                # after si-tests check finished, we may check expected reports from lcm-agent,
                # about actial information from node, but in perspective of lcm-agent. PRODX-22958 might be related.
                host_info = \
                    self.cluster.get_cluster_lcmmachine(name=m.name,
                                                        namespace=self.cluster.namespace).data['status']['hostInfo']
                assert host_info['kernelVersion'] in test_allowed_distr_kernel, \
                    f"Expected machine=kernel on host os: {m.name} in {test_allowed_distr_kernel}," \
                    f"but lcm-agent report:\n {host_info['kernelVersion']}"
                return

    def check_created_machine(self, machine: "Machine"):
        LOG.info(f'Check newly created machine {machine.name}')

        # Waiting for master node become Ready
        self.cluster.check.wait_machine_status_by_name(machine_name=machine.name,
                                                       expected_status='Ready')

        # Waiting for machines are Ready
        self.cluster.check.check_machines_status(timeout=1800)
        self.cluster.check.check_cluster_nodes()
        self.cluster.check.check_k8s_nodes()

        # Check/wait for correct docker service replicas in cluster
        ucp_worker_agent_name = self.cluster.check.get_ucp_worker_agent_name()
        self.cluster.check.check_actual_expected_docker_services(
            changed_after_upd={'ucp-worker-agent-x': ucp_worker_agent_name})
        self.cluster.check.check_k8s_pods()
        self.cluster.check.check_actual_expected_pods(timeout=3200)
        self.cluster.check.check_cluster_readiness()

    def check_deleted_node(self, node_name: str):
        LOG.info(f"Wait k8s node delete - {node_name}")
        waiters.wait(lambda: node_name not in self.cluster.get_k8s_node_names(),
                     interval=10, timeout=1800)

        self.cluster.check.check_machines_status()
        self.cluster.check.check_cluster_nodes()
        self.cluster.check.check_k8s_nodes()

        left_machines = self.cluster.get_machines_names()

        LOG.info("Check number of k8s nodes")
        left_nodes = self.cluster.get_k8s_node_names()

        assert len(left_machines) == len(left_nodes), \
            f"Unequal number of nodes and machines\nnodes: {left_nodes}\nmachines: {left_machines}"

    def check_overlay_encryption_functionality(self):
        if not self.cluster.data['spec'].get('providerSpec', {}).get('value', {}).get('secureOverlay', False):
            LOG.info('secureOverlay disabled. Skip encryption checking')
        else:
            # TODO(akasatkin): add test to verity Wireguard (see PRODX-29891)
            LOG.info('secureOverlay enabled. Test will be implemented in PRODX-29891')
            # self.cluster.check.check_ipsec_functionality()

    def check_ipsec_functionality(self):
        # 1. Check ipsec policy
        # 2. Check ipsec encryption for tunnel traffic
        if not self.cluster.data['spec'].get('providerSpec', {}).get('value', {}).get('secureOverlay', False):
            LOG.info('secureOverlay disabled. Skip ipsec checking')
        else:
            machines = self.cluster.get_machines()
            err_msg = ''
            for machine in machines:
                neighbor_machines_ips = [m.internal_ip for m in machines if m.internal_ip != machine.internal_ip]
                ipsec_policy_cmd = 'sudo ip xfrm policy list|grep " ipencap"|cut -d" " -f4|cut -d"/" -f1'
                tcpdump_cmd = "timeout 6 tcpdump -nei any esp or ah -c300 -q 2>/dev/null"\
                              "|sed 's/  In/ I/g'|cut -d' ' -f4,6|tr -d ':'"
                policy_result = machine.exec_pod_cmd(ipsec_policy_cmd, verbose=False)
                tcpdump_result = machine.exec_pod_cmd(tcpdump_cmd, verbose=False)
                policy_list = policy_result['logs'].strip().split('\n')
                tcpdump_data = tcpdump_result['logs'].strip().split('\n')
                LOG.info(f"Сheck ipsec policy on machine {machine.name}/{machine.internal_ip}")
                if policy_result['exit_code'] != 0:
                    err_msg += f"Ipsec policy check error {self._cluster.name}/{machine.name}:"\
                               f"\n{policy_result['logs']}\n"
                for neighbor_ip in neighbor_machines_ips:
                    if neighbor_ip not in policy_list:
                        err_msg += f"Ipsec disabled for tunnel between {machine.name} and dst tunnel ip {neighbor_ip}\n"
                LOG.info(f"Сheck ipsec traffic on machine {machine.name}/{machine.internal_ip}")
                if tcpdump_result['exit_code'] != 0:
                    err_msg += f"Ipsec traffic check error {self._cluster.name}/{machine.name}:"\
                               f"\n{tcpdump_result['logs']}\n"
                tcpdump_neighbors = []
                if tcpdump_result['logs'] == "\n" or len(tcpdump_data) == 0:
                    err_msg += f"Ipsec traffic not found in 6 second on {machine.name}:\n{tcpdump_result['logs']}\n"
                else:
                    for src_dst in set(tcpdump_data):
                        addresses = src_dst.split(' ')
                        if len(addresses) != 2:
                            LOG.info("Unexpected count of addresses. There must be 2 addresses (src and dst)")
                        else:
                            src, dst = addresses
                            if src == machine.internal_ip:
                                tcpdump_neighbors.append(dst)
                            else:
                                tcpdump_neighbors.append(src)
                for neighbor_ip in neighbor_machines_ips:
                    if neighbor_ip not in tcpdump_neighbors:
                        err_msg += f"Ipsec traffic is not detected between src machine {machine.name} "\
                                   f"and dst machine ip {neighbor_ip}\n"
            assert err_msg == '', err_msg

    def check_expected_actual_reboot_required_status(self):
        kaas_version = self.cluster.get_kaasrelease_version()
        cluster_version = self.cluster.clusterrelease_version
        if not version.parse(kaas_version) > version.parse("kaas-2-19-0"):
            LOG.info(f"KaaS version {kaas_version} is not supporting reboot required status for machines")
            return
        if 'mosk' in cluster_version and not version.parse(cluster_version) > version.parse('mosk-8-8-0-22-3'):
            LOG.info(f"Reboot require status for machines is not supported for clusterrelease "
                     f"{cluster_version}")
            return
        LOG.info("Compare reboot required status between KaaS machines and hosts")
        machines = self.cluster.get_machines()
        failed_machines = {'failed_reboot_status': {}, 'failed_packages_reason': {}}
        for machine in machines:
            m_name = machine.name
            actual_reboot_required_status_from_machine = machine.is_reboot_required()
            expected_reboot_required_status_from_host = machine.reboot_required_status_from_host()
            actual_status_from_machine = actual_reboot_required_status_from_machine[0]
            expected_status_from_host = expected_reboot_required_status_from_host[0]
            actual_packges_reason = actual_reboot_required_status_from_machine[1]
            expected_packges_reason = expected_reboot_required_status_from_host[1]
            LOG.info(f"Current reboot required status for machine {m_name} is "
                     f"{actual_status_from_machine}, expected: {expected_status_from_host}. Actual packages reason: "
                     f"{actual_packges_reason}, expected: {expected_packges_reason}")
            if actual_status_from_machine != expected_status_from_host:
                failed_machines['failed_reboot_status'].update(
                    {m_name: {'expected_reboot_required_status': expected_status_from_host,
                              'actual_reboot_required_status': actual_status_from_machine}})
            if actual_packges_reason != expected_packges_reason:
                failed_machines['failed_packages_reason'].update(
                    {m_name: {'expected_packages_reason': expected_packges_reason,
                              'actual_packages_reason': actual_packges_reason}})
        fail_message = ""
        failed_reboot_status = failed_machines.get('failed_reboot_status')
        failed_packages_reason = failed_machines.get('failed_packages_reason')
        if failed_reboot_status:
            fail_message += (f"\n\nReboot status mismatch for next machines:\n"
                             f"{yaml.dump(failed_reboot_status)}")
        if failed_packages_reason:
            fail_message += (f"\nPackages status mismatch for next machines:\n"
                             f"{yaml.dump(failed_packages_reason)}")

        assert not fail_message, fail_message
        LOG.info("Reboot required status and reason packages for all machines as expected")

    def check_connection_via_bastion(self, target_machine: "Machine"):
        """
        Check connection to target machine internal IP via cluster bastion host
        Args:
            target_machine (Machine): target machine object
        Returns: None

        """
        expected_name = target_machine.get_k8s_node_name()
        destination_ip = target_machine.internal_ip

        LOG.info(f"Trying to connect to node {expected_name} with IP {destination_ip} "
                 f"via bastion host with IP {self.cluster.bastion_ip}")
        keys = utils.load_keyfile(self.cluster.private_key)
        key = utils.get_rsa_key(keys['private'])
        auth = exec_helpers.SSHAuth(username=settings.KAAS_BASTION_SSH_LOGIN,
                                    password='', key=key)
        ssh = self.cluster.get_bastion_remote()
        result = ssh.execute_through_host(destination_ip, "hostname", auth=auth)
        assert result.stdout_str == expected_name

    def check_os_vm(self, openstack_client, image, flavor, az_name,
                    metadata=None, volume_size=None, boot_from_volume=False, vm_name=None, vm_id=None):
        """
        Check OpenStack VM parameters
        Args:
            openstack_client: OpenStack client
            vm_name: OpenStack instance name (instead of vm_id)
            vm_id: OpenStack instance id (instead of vm_name)
            image: OpenStack image name
            flavor: OpenStack flavor name
            az_name: Availability Zone name
            metadata: expected metadata
            boot_from_volume: allowed boot from volume flag
            volume_size: expected volume size

        Returns: None

        """
        LOG.info("Check OpenStack VM parameters")
        if vm_name:
            openstack_machine = openstack_client.get_server_by_name(vm_name).to_dict()
        elif vm_id:
            openstack_machine = openstack_client.get_server_by_id(vm_id).to_dict()
        else:
            raise Exception("vm_name or vm_id should be specified")
        os_metadata = openstack_machine.get('metadata')
        os_flavor_id = openstack_machine.get('flavor', {}).get('id')
        os_flavor = openstack_client.nova.flavors.get(flavor=os_flavor_id)
        os_image = openstack_machine.get('image')
        os_az_name = openstack_machine.get('OS-EXT-AZ:availability_zone')
        assert os_flavor.name == flavor, "VM is created from incorrect flavor"
        assert os_az_name == az_name
        # check VM metadata
        error_keys = {}
        if metadata:
            for key, value in metadata.items():
                if (v := os_metadata.get(key)) and v == value:
                    pass
                else:
                    error_keys[key] = value
        assert not error_keys, "The following metadata has not been populated {0}".format(error_keys)

        if boot_from_volume:
            LOG.info("Check OpenStack VM boot from volume parameters")
            os_attached_volume = openstack_machine.get('os-extended-volumes:volumes_attached')
            os_volume = os_attached_volume[0].get('id')
            cinder_volume = openstack_client.cinder.volumes.get(volume_id=os_volume)
            cinder_image_name = cinder_volume.volume_image_metadata.get('image_name')
            assert os_image == '', 'Should be empty if booted from volume'
            assert len(os_attached_volume) == 1, "Only one attached volume should be exist"
            assert cinder_volume.size == volume_size, "Volume size is incorrect"
            assert cinder_volume.bootable == 'true', "Volume should be bootable for boot from volume feature"
            assert cinder_image_name == image, "Volume created from incorrect image"
        else:
            # VM image parameter exist only if deploy without boot from volume option
            image_id = os_image.get('id')
            glance_image = openstack_client.nova.glance.find_image(image_id)
            assert glance_image.name == image, "VM is created from incorrect image"

    def check_aws_vm(self, instance_id, instance_type, ami_id):
        """
        Check AWS VM parameters

        Args:
            instance_id: AWS instance ID
            instance_type: Type of instance used to launch the bastion instance.
            ami_id: AMI ID used to launch the bastion instance.

        Returns: None

        """
        LOG.info(f"Check AWS VM parameters for Instance ID = {instance_id}")
        aws_client = self.cluster.provider_resources.client
        aws_instance = aws_client.get_instance(instance_id=instance_id)
        assert instance_type == aws_instance.instance_type, "VM is created from incorrect instance type"
        assert ami_id == aws_instance.image.image_id, "VM is created from incorrect AMI"

    def check_deploy_stage_success(self, skipped_stages_names=None):
        skipped_stages = []
        if skipped_stages_names and type(skipped_stages_names) is str:
            skipped_stages.append(skipped_stages_names)
        elif skipped_stages_names and type(skipped_stages_names) is list:
            skipped_stages.extend(skipped_stages_names)
        LOG.info(f"Check stages in crd clusterdeploy for cluster {self.cluster.name}")
        failed_stages = []

        def _inner():
            nonlocal failed_stages
            failed_stages = []
            stages = self.cluster.get_cluster_deploy_stages_filtered_by_cluster_name(self.cluster.name)
            if stages:
                for stage in stages:
                    stage_name = stage.get('name', '')
                    timestamp = stage.get('timestamp', '')
                    message = stage.get('message', '')
                    success = stage.get('success', '')
                    if not success:
                        if stage_name not in skipped_stages:
                            failed_stages.append(f"Cluster: {self.cluster.name}; "
                                                 f"Failed stage: {stage_name}; "
                                                 f"Success: {success}; "
                                                 f"Message: {message}; "
                                                 f"Time of last retry: {timestamp} ")
            return len(failed_stages) == 0

        try:
            waiters.wait(_inner)
        except exceptions.TimeoutError:
            pass
        assert len(failed_stages) == 0, f'Next stages are failed {failed_stages}'

    def check_upgrade_stage_success(self, skipped_stages_names=None):
        # TODO(tleontovich): Delete can_check_mke after 2.22 is out
        # TODO(tleontovich): Delete can_check_mos after 2.23.1 is out
        skipped_stages = []
        if skipped_stages_names:
            skipped_stages.append(skipped_stages_names)
        cluster_release = self.cluster.clusterrelease_version
        LOG.info(f"Cluster release: {cluster_release}")
        can_check_mke_11 = 'mke-11' in cluster_release and version.parse(
            cluster_release) >= version.parse('mke-11-6-0-rc-3-5-5')
        can_check_mke_7 = 'mke-7' in cluster_release and version.parse(
            cluster_release) >= version.parse('mke-7-12-0-rc-3-4-11')
        can_check_mos = 'mos' in cluster_release and version.parse(
            cluster_release) >= version.parse('mosk-12-6-0-rc-23-1')

        if can_check_mke_7 or can_check_mke_11 or can_check_mos:
            failed_stages = []
            LOG.info(f"Check stages in crd clusterupgradehistroy "
                     f"for cluster {self.cluster.name}")
            stages = self.cluster.get_cluster_upgrade_stages_filtered_by_cluster_name(
                self.cluster.name)
            if stages:
                for stage in stages:
                    stage_name = stage.get('name', '')
                    timestamp = stage.get('timestamp', '')
                    message = stage.get('message', '')
                    success = stage.get('success', '')
                    status = stage.get('status', '')
                    if not success and status != 'NotStarted':
                        if stage_name not in skipped_stages:
                            failed_stages.append(f"Cluster: {self.cluster.name}; "
                                                 f"Failed stage: {stage_name}; "
                                                 f"Status: {status}; "
                                                 f"Success: {success}; "
                                                 f"Message: {message}; "
                                                 f"Time of last retry: {timestamp} ")
                assert len(failed_stages) == 0, f'Next stages are failed {failed_stages}'

    def check_upgraded_machines_not_contain_cordon_drain_stages(self, status_matches=None):
        """
        For patch releases we expect that the cordon drain stages did not occur.

        All statuses here: https://gerrit.mcp.mirantis.com/plugins/gitiles/kaas/core/+/
        refs/heads/master/vendor/github.com/Mirantis/kaas-api/v2/pkg/apis/lcm/v1alpha1/lcmstatus_types.go

        Args:
            status_matches = [] - List of matches to check in statuses
        """
        LOG.info("Check that cordon/drain stages after the nodes upgrade did not run")
        if not status_matches:
            status_matches = ['Kubernetes drained']
        elif status_matches and isinstance(status_matches, str):
            status_matches = [status_matches]
        stages_present = []
        for m in self.cluster.get_machines():
            LOG.info(f"Check stages in crd machineupgradehistroy for machine {m.name}")
            stages = self.cluster.get_machine_upgrade_stages_filtered_by_machine_name(machine_name=m.name)
            if stages:
                for match in status_matches:
                    for stage in stages:
                        stage_name = stage.get('name', '')
                        timestamp = stage.get('timestamp', '')
                        message = stage.get('message', '')
                        success = stage.get('success', '')
                        if match in stage_name:
                            stages_present.append(f"Cluster: {self.cluster.name}; "
                                                  f"Machine: {m.name}; "
                                                  f"Stage: {stage_name}; "
                                                  f"Success: {success}; "
                                                  f"Message: {message}; "
                                                  f"Time of last retry: {timestamp} ")
        assert len(stages_present) == 0, f'Next stages are present, but should not: {stages_present}'

    def wait_leader_migrate_from_node(self, node_name, timeout=600):
        """
        Wait until the leader migrates from the node during the election process
        Args:
            node_name: target node that pods should migrate from
            timeout: wait timeout in seconds

        Returns: None

        """
        pods_on_node = self.cluster.get_pods_for_node(node_name=node_name)
        pods_name = [p.name for p in pods_on_node]

        def _get_data():
            leader: dict = self.cluster.get_leader_pods()
            filtered = {k: v for k, v in leader.items() if v in pods_name}
            LOG.info(f"Next leader pods still not changed: {filtered}")
            return False if filtered else True

        LOG.info(f"Wait until leaders migrate from node {node_name}")
        waiters.wait(_get_data, interval=30, timeout=timeout)

    def check_auditd(self):
        """
        Scenario:
            1. Get machines where auditd is running.
               NOTE: All machines of a cluster should have auditd
                     running if it was enabled for a cluster.
               Use `exec_pod_cmd` to issue check commands.
            2. Check if auditd service is running.
            3. Check if auditd preset rules are present.
            4. Check if auditd custom rules are present.
            5. Check if auditd log rotation crontask is installed.
               If it is, try to execute it and print the list of log
               files before and after the rotation.

        Note: Some checks did not verify correctness of the auditd
              state, e.g. if log rotation crontask doesn't exist it
              is not considered as a failure. This is due to
              optional nature of such states, and due to it's not possible
              to verify them without getting the desired state from the
              cluster object, what is an overkill now.
        """
        machines = self.cluster.get_machines()
        for machine in machines:
            LOG.info(f"Accessing {machine.name}")

            LOG.info("Checking auditd.service ...")
            cmd = machine.exec_pod_cmd("systemctl is-active auditd.service")
            assert cmd['exit_code'] == 0, \
                f"auditd.service is not active on {machine.name}:\n{cmd['logs']}"

            LOG.info("Looking for auditd preset rules ...")
            cmd = machine.exec_pod_cmd("ls -1 /etc/audit/rules.d")
            LOG.info(f"Found the following auditd rules: {cmd['logs']}")

            LOG.info("Looking for auditd custom rules ...")
            custom_rules_files = [
                '60-custom.rules',
                '60-custom-x32.rules',
                '60-custom-x64.rules'
            ]
            for filename in custom_rules_files:
                cmd = machine.exec_pod_cmd(f"cat /etc/audit/rules.d/{filename}")

                if cmd['exit_code'] == 0:
                    LOG.info(f"Custom rules file '{filename}' contains:\n{cmd['log']}")
                elif cmd['exit_code'] == 1:
                    LOG.info(f"Seems that file '{filename}' does not exist.")
                    continue
                else:
                    LOG.error(f"Unexpected error '{cmd['exit_code']}'")

            LOG.info("Checking auditd log rotation ...")
            cmd = machine.exec_pod_cmd("test -f /etc/cron.daily/auditd.cron")

            if cmd['exit_code'] == 0:
                LOG.info("Auditd crontask installed.")
                auditd_cron_exists = True
            elif cmd['exit_code'] == 1:
                LOG.info("Auditd crontask not installed.")
                auditd_cron_exists = False
            else:
                raise Exception(f"Unexpected exit code '{cmd['exit_code']}'")

            if auditd_cron_exists:
                cmd = machine.exec_pod_cmd("ls -1 /var/log/audit")
                LOG.info(f"auditd log files before rotation:\n{cmd['logs']}")

                cmd = machine.exec_pod_cmd("sh /etc/cron.daily/auditd.cron")
                assert cmd['exit_code'] == 0, \
                    f"Failed to execute auditd.cron on {machine.name}:\n{cmd['logs']}"

                cmd = machine.exec_pod_cmd("ls -1 /var/log/audit")
                LOG.info(f"auditd log files after rotation:\n{cmd['logs']}")
