from hashlib import sha256
import multiprocessing

from si_tests.deployments.deploy_refapp import Refapp
from si_tests.deployments.utils import commons
from si_tests.utils import waiters

NMR_META = {
    "group": "lcm.mirantis.com",
    "version": "v1alpha1",
    "plural": "nodemaintenancerequests",
}

NWL_META = {
    "group": "lcm.mirantis.com",
    "version": "v1alpha1",
    "plural": "nodeworkloadlocks",
}


def host_hash(host):
    host = host.split(".")[0]
    return sha256(host.encode()).hexdigest()[:8]


def create_node_maintenance_request(api, node, name):
    nmr_body = {
        "apiVersion": "lcm.mirantis.com/v1alpha1",
        "kind": "NodeMaintenanceRequest",
        "metadata": {"name": name},
        "spec": {"nodeName": node, "scope": "os"},
    }
    api.create_cluster_custom_object(**NMR_META, body=nmr_body)


def delete_node_maintenance_request(api, name):
    api.delete_cluster_custom_object(**NMR_META, name=name, body={})


def get_node_workload_lock_status(api, name):
    nwl = api.get_cluster_custom_object_status(**NWL_META, name=name)
    return nwl["status"]["state"]


def check_node_controller(openstack_client_manager, os_manager):

    hostname_attr = 'OS-EXT-SRV-ATTR:hypervisor_hostname'

    hypervisor_list = openstack_client_manager.hypervisor.list([])

    assert len(hypervisor_list) > 1, ("Environment has 1 or less "
                                      "compute nodes %s", hypervisor_list)

    for hypervisor in hypervisor_list:
        hostname = hypervisor['Hypervisor Hostname']
        commons.LOG.info("Start maintenance mode on %s", hostname)
        nmr_name = "nmr-si-test-" + host_hash(hostname)
        nwl_name = "openstack-" + hostname.split('.')[0]
        create_node_maintenance_request(
            os_manager.api.api_custom, hostname, nmr_name)

        # Check evacuation finished successfully
        commons.LOG.info("Check evacuation finished successfully")

        def _wait_os_workload_status(status='inactive'):
            return (get_node_workload_lock_status(
                os_manager.api.api_custom, nwl_name) == status)

        waiters.wait(_wait_os_workload_status,
                     timeout=1800,
                     interval=30)
        commons.LOG.info("Check node controller evacuation "
                         "finished successfully.")

        def _check_instances_migrated(hostname):
            instances = openstack_client_manager.server.list(['--all'])
            instances = [x['ID'] for x in instances]
            not_migrated = []
            for instance_id in instances:
                instance = openstack_client_manager.server.show(
                    [instance_id])
                if instance[hostname_attr] == hostname:
                    not_migrated.append(instance['id'])
            if not_migrated:
                msg = "Some instances not migrated yet: %s"
                commons.LOG.info(msg, not_migrated)
                return False
            return True

        waiters.wait(_check_instances_migrated,
                     predicate_args=(hostname,),
                     timeout=1800,
                     interval=30)

        commons.LOG.info("Return node from maintenance mode to status mode.")
        delete_node_maintenance_request(os_manager.api.api_custom, nmr_name)
        waiters.wait(_wait_os_workload_status,
                     predicate_args=('active',),
                     timeout=1800,
                     interval=30)


def test_node_controller_evacuate(openstack_client_manager, os_manager):
    """ Test node controller evacuation
      1) Deploy RefApp
      2) Run in parallel refapp requests and
         maintenance mode for node
      3) Check that evacuation finished successfully
    """
    refapp = Refapp(openstack_client_manager)
    refapp.deploy()

    commons.LOG.info("Init 2 processes check controller evacuation "
                     "and RefApp pings")
    process_check_controller = multiprocessing.Process(
        target=check_node_controller,
        args=(openstack_client_manager, os_manager))

    process_collect_statistic = multiprocessing.Process(
        target=refapp.collect_statistic)

    commons.LOG.info("Start Check controller evacuation and "
                     "refapp ping processes")
    process_check_controller.start()
    process_collect_statistic.start()

    commons.LOG.info("Wait when Check node evacuation process "
                     "finished")
    process_check_controller.join()

    commons.LOG.info("Terminate refapp pings process")
    process_collect_statistic.terminate()
    process_collect_statistic.join()

    exitcode = process_check_controller.exitcode
    assert not exitcode, (
        "Test Node controller evacuate failed with "
        "exit code: %s", exitcode)
