import yaml
from yaml import scanner as yaml_scanner
from si_tests.utils import utils, waiters
from si_tests import logger, settings

LOG = logger.logger


def test_tf_backup_resotre(kaas_manager, show_step):
    """ Automatically back up and restore TF data

    Scenario:
        1. Get openstack client pod
        2. Create network before backup for comparing in database between backup and restore
        3. Enable backup job via TF operator CR
        4. Wait for backup job done
        5. Disable backup job via TF operator CR
        6. Attach network tag after backup for comparing in database between backup and restore
        7. Enable restore via TF operator CR
        8. Wait for restore process to be completed
        9. Wait for contrail api readiness
        10. Check for network exist
        11. Check for tags is not exist
    """

    cluster_name = settings.TARGET_CLUSTER
    namespace_name = settings.TARGET_NAMESPACE
    operator_name = 'openstack-tf'
    operator_ns = 'tf'
    ns = kaas_manager.get_namespace(namespace_name)
    child_cluster = ns.get_cluster(cluster_name)

    client_pods = child_cluster.k8sclient.pods.list(namespace="openstack", name_prefix='keystone-client')
    show_step(1)
    assert len(client_pods) > 0, "No pods found with prefix keystone-client in namespace openstack"
    client_pod = client_pods[0]
    network_name = 'tf-database-test-net'
    tag_name = 'tf-database-test-tag'

    # Create network before backup
    cmd_create_net = ['/bin/sh', '-c', f'openstack network create {network_name}']
    show_step(2)
    client_pod.exec(cmd_create_net)

    tfoperator = child_cluster.k8sclient.tfoperator.get(name=operator_name, namespace=operator_ns)
    LOG.info(f"tf operator name is {tfoperator.name}")
    show_step(3)
    tfoperator.patch({"spec":
                      {"controllers":
                       {"tf-dbBackup":
                        {"enabled": True}}}})

    backup_jobs_prefix = "tf-dbbackup-job-"
    waiters.wait(lambda: len([cj for cj in child_cluster.k8sclient.cronjobs.list(name_prefix="tf-dbbackup-job",
                                                                                 namespace=operator_ns)]) > 0,
                 timeout=60, interval=10, timeout_msg="Timeout for waiting cron job")
    backup_cron_job = child_cluster.k8sclient.cronjobs.list(name_prefix="tf-dbbackup-job", namespace=operator_ns)[0]
    job = backup_cron_job.read()
    job_template = job.spec.job_template
    job_template.metadata.name = backup_jobs_prefix + utils.gen_random_string(3)
    child_cluster.k8sclient.jobs.create(body=job_template, namespace=operator_ns)
    show_step(4)

    waiters.wait(lambda: len([j for j in child_cluster.k8sclient.jobs.list(name_prefix=backup_jobs_prefix,
                                                                           namespace=operator_ns)]) > 0,
                 timeout=60, interval=10, timeout_msg="Timeout for waiting job")

    child_cluster.k8sclient.jobs.check_jobs_completed(target_namespaces=operator_ns, jobs_prefix=backup_jobs_prefix)
    show_step(5)
    tfoperator.patch({"spec":
                      {"controllers":
                       {"tf-dbBackup":
                        {"enabled": False}}}})

    # Attach tag for network after the backup
    cmd_add_tag_for_net = ['/bin/sh', '-c', f'openstack network set {network_name} --tag {tag_name}']
    show_step(6)
    client_pod.exec(cmd_add_tag_for_net)

    show_step(7)
    tfoperator.patch({"spec":
                      {"settings":
                       {"dbRestoreMode":
                        {"enabled": True}}}})

    # Wait for restore process to be completed
    show_step(8)
    waiters.wait(lambda: len([c for c in child_cluster.k8sclient.tfdbrestores.list(namespace=operator_ns)]) > 0,
                 timeout=600, interval=10,  timeout_msg="Timeout wait for tfdbrestore object")
    tfdbrestore = child_cluster.k8sclient.tfdbrestores.list(namespace=operator_ns)[0]
    waiters.wait(lambda: tfdbrestore.data['status']['health'] == "Ready",
                 timeout=1200, interval=30, timeout_msg="Timeout wait for restoration status Ready")

    # Wait for contrail api readiness
    show_step(9)
    cmd_net_list = ['/bin/sh', '-c', 'openstack network list -f yaml']

    def check_contrail_api_readiness():
        try:
            yaml.safe_load(client_pod.exec(cmd_net_list))
            return True
        except yaml_scanner.ScannerError as e:
            LOG.warning(f"Got error {e} during parsing command status'")
            return False
    waiters.wait(check_contrail_api_readiness,
                 timeout=150, interval=15, timeout_msg="Timeout: Contrail api is not Ready")

    # Check for network exist
    show_step(10)
    network_list = yaml.safe_load(client_pod.exec(cmd_net_list))
    network_counter = sum(n.get('Name') == network_name for n in network_list)
    assert network_counter == 1, "Created network before db backups is not exist after the restore"

    # Check for tags is not exist
    cmd_tag_in_net = ['/bin/sh', '-c', f'openstack network show {network_name} -c tags -f yaml']
    show_step(11)
    tags_list = yaml.safe_load(client_pod.exec(cmd_tag_in_net))
    tag_counter = sum(t == tag_name for t in tags_list['tags'])
    assert tag_counter == 0, "Tags should not be exist after restore"
